#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import grpc

from skywalking import config
from skywalking.client import ServiceManagementClient, TraceSegmentReportService, ProfileTaskChannelService, \
    LogDataReportService
from skywalking.command import command_service
from skywalking.loggings import logger, logger_debug_enabled
from skywalking.profile import profile_task_execution_service
from skywalking.profile.profile_task import ProfileTask
from skywalking.protocol.common.Common_pb2 import KeyStringValuePair
from skywalking.protocol.language_agent.Tracing_pb2_grpc import TraceSegmentReportServiceStub
from skywalking.protocol.logging.Logging_pb2_grpc import LogReportServiceStub
from skywalking.protocol.management.Management_pb2 import InstancePingPkg, InstanceProperties
from skywalking.protocol.management.Management_pb2_grpc import ManagementServiceStub
from skywalking.protocol.profile.Profile_pb2 import ProfileTaskCommandQuery, ProfileTaskFinishReport
from skywalking.protocol.profile.Profile_pb2_grpc import ProfileTaskStub


class GrpcServiceManagementClient(ServiceManagementClient):
    def __init__(self, channel: grpc.Channel):
        self.service_stub = ManagementServiceStub(channel)

    def send_instance_props(self):
        self.service_stub.reportInstanceProperties(InstanceProperties(
            service=config.service_name,
            serviceInstance=config.service_instance,
            properties=[KeyStringValuePair(key='language', value='Python')],
        ))

    def send_heart_beat(self):
        if logger_debug_enabled:
            logger.debug(
                'service heart beats, [%s], [%s]',
                config.service_name,
                config.service_instance,
            )
        self.service_stub.keepAlive(InstancePingPkg(
            service=config.service_name,
            serviceInstance=config.service_instance,
        ))


class GrpcTraceSegmentReportService(TraceSegmentReportService):
    def __init__(self, channel: grpc.Channel):
        self.report_stub = TraceSegmentReportServiceStub(channel)

    def report(self, generator):
        self.report_stub.collect(generator)


class GrpcLogDataReportService(LogDataReportService):
    def __init__(self, channel: grpc.Channel):
        self.report_stub = LogReportServiceStub(channel)

    def report(self, generator):
        self.report_stub.collect(generator)


class GrpcProfileTaskChannelService(ProfileTaskChannelService):
    def __init__(self, channel: grpc.Channel):
        self.profile_stub = ProfileTaskStub(channel)

    def do_query(self):
        query = ProfileTaskCommandQuery(
            service=config.service_name,
            serviceInstance=config.service_instance,
            lastCommandTime=profile_task_execution_service.get_last_command_create_time()
        )

        commands = self.profile_stub.getProfileTaskCommands(query)
        command_service.receive_command(commands)

    def send(self, generator):
        self.profile_stub.collectSnapshot(generator)

    def finish(self, task: ProfileTask):
        finish_report = ProfileTaskFinishReport(
            service=config.service_name,
            serviceInstance=config.service_instance,
            taskId=task.task_id
        )
        self.profile_stub.reportTaskFinish(finish_report)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context, NoopContext
from skywalking.trace.span import NoopSpan
from skywalking.trace.tags import TagHttpMethod, TagHttpURL, TagHttpStatusCode

link_vector = ['https://docs.aiohttp.org']
support_matrix = {
    'aiohttp': {
        '>=3.10': [],  # waiting for 3.8 release
        '>=3.6': ['3.7.4']
    }
}
note = """"""


def install():
    from aiohttp import ClientSession
    from aiohttp.web_protocol import RequestHandler
    from multidict import CIMultiDict, MultiDict, MultiDictProxy
    from yarl import URL

    async def _sw_request(self: ClientSession, method: str, str_or_url, **kwargs):
        url = URL(str_or_url).with_user(None).with_password(None)
        peer = f"{url.host or ''}:{url.port or ''}"

        span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
            else get_context().new_exit_span(op=url.path or '/', peer=peer, component=Component.AioHttp)

        with span:
            span.layer = Layer.Http
            span.tag(TagHttpMethod(method.upper()))  # pyre-ignore
            span.tag(TagHttpURL(str(url.with_password(None))))  # pyre-ignore

            carrier = span.inject()
            headers = kwargs.get('headers')

            if headers is None:
                headers = kwargs['headers'] = CIMultiDict()
            elif not isinstance(headers, (MultiDictProxy, MultiDict)):
                headers = CIMultiDict(headers)

            for item in carrier:
                headers.add(item.key, item.val)

            res = await _request(self, method, str_or_url, **kwargs)

            span.tag(TagHttpStatusCode(res.status))

            if res.status >= 400:
                span.error_occurred = True

            return res

    _request = ClientSession._request
    ClientSession._request = _sw_request

    async def _sw_handle_request(self, request, start_time: float):
        carrier = Carrier()
        method = request.method

        for item in carrier:
            val = request.headers.get(item.key)

            if val is not None:
                item.val = val

        span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
            else get_context().new_entry_span(op=request.path, carrier=carrier)

        with span:
            span.layer = Layer.Http
            span.component = Component.AioHttp
            peer_name = request._transport_peername
            if isinstance(peer_name, (list, tuple)):
                span.peer = f'{peer_name[0]}:{peer_name[1]}'
            else:
                span.peer = f'{peer_name}'

            span.tag(TagHttpMethod(method))  # pyre-ignore
            span.tag(TagHttpURL(str(request.url)))  # pyre-ignore

            resp, reset = await _handle_request(self, request, start_time)

            span.tag(TagHttpStatusCode(resp.status))

            if resp.status >= 400:
                span.error_occurred = True

        return resp, reset

    _handle_request = RequestHandler._handle_request
    RequestHandler._handle_request = _sw_handle_request



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ..common import Common_pb2 as common_dot_Common__pb2
from ..logging import Logging_pb2 as logging_dot_Logging__pb2


class LogReportServiceStub(object):
    """Report collected logs into the OAP backend
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.collect = channel.stream_unary(
                '/skywalking.v3.LogReportService/collect',
                request_serializer=logging_dot_Logging__pb2.LogData.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )


class LogReportServiceServicer(object):
    """Report collected logs into the OAP backend
    """

    def collect(self, request_iterator, context):
        """Recommend to report log data in a stream mode.
        The service/instance/endpoint of the log could share the previous value if they are not set.
        Reporting the logs of same service in the batch mode could reduce the network cost.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_LogReportServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'collect': grpc.stream_unary_rpc_method_handler(
                    servicer.collect,
                    request_deserializer=logging_dot_Logging__pb2.LogData.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'skywalking.v3.LogReportService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class LogReportService(object):
    """Report collected logs into the OAP backend
    """

    @staticmethod
    def collect(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/skywalking.v3.LogReportService/collect',
            logging_dot_Logging__pb2.LogData.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Component, agent, config
from skywalking import profile
from skywalking.agent import isfull
from skywalking.profile.profile_status import ProfileStatusReference
from skywalking.trace import ID
from skywalking.trace.carrier import Carrier
from skywalking.trace.segment import Segment, SegmentRef
from skywalking.trace.snapshot import Snapshot
from skywalking.trace.span import Span, Kind, NoopSpan, EntrySpan, ExitSpan
from skywalking.utils.counter import Counter
from skywalking.utils.time import current_milli_time

try:  # attempt to use async-local instead of thread-local context and spans
    import contextvars

    __spans = contextvars.ContextVar('spans')
    _spans = __spans.get
    _spans_set = __spans.set  # pyre-ignore

    def _spans():  # need to do this because can't set mutable default = [] in contextvars.ContextVar()
        spans = __spans.get(None)

        if spans is not None:
            return spans

        spans = []
        __spans.set(spans)

        return spans

    def _spans_dup():
        spans = __spans.get()[:]
        __spans.set(spans)

        return spans

    __spans.set([])

except ImportError:
    import threading

    class SwLocal(threading.local):
        def __init__(self):
            self.spans = []

    __local = SwLocal()

    def _spans():
        return __local.spans

    def _spans_set(spans):
        __local.spans = spans

    _spans_dup = _spans


class SpanContext(object):
    def __init__(self):
        self.segment = Segment()  # type: Segment
        self._sid = Counter()
        self._correlation = {}  # type: dict
        self._nspans = 0
        self.profile_status = None  # type: ProfileStatusReference
        self.create_time = current_milli_time()

    def ignore_check(self, op: str, kind: Kind, carrier: 'Carrier' = None):
        if config.RE_IGNORE_PATH.match(op) or isfull() or (carrier is not None and carrier.is_suppressed):
            return NoopSpan(context=NoopContext())

        return None

    def new_span(self, parent: Span, SpanType: type, **kwargs) -> Span: # noqa
        finished = parent and not parent._depth
        context = SpanContext() if finished else self
        span = SpanType(context=context,
                        sid=context._sid.next(),
                        pid=parent.sid if parent and not finished else -1,
                        **kwargs)

        # if parent finished and segment was archived before this child starts then we need to refer to parent
        if finished:
            carrier = Carrier(
                trace_id=str(parent.context.segment.related_traces[0]),
                segment_id=str(parent.context.segment.segment_id),
                span_id=str(parent.sid),
                service=config.service_name,
                service_instance=config.service_instance,
                endpoint=parent.op,
                client_address=parent.peer,
                correlation=parent.context._correlation,
            )

            Span.extract(span, carrier)

        return span

    def new_local_span(self, op: str) -> Span:
        span = self.ignore_check(op, Kind.Local)
        if span is not None:
            return span

        spans = _spans()
        parent = spans[-1] if spans else None  # type: Span

        return self.new_span(parent, Span, op=op, kind=Kind.Local)

    def new_entry_span(self, op: str, carrier: 'Carrier' = None, inherit: Component = None) -> Span:
        span = self.ignore_check(op, Kind.Entry, carrier)
        if span is not None:
            return span

        spans = _spans()
        parent = spans[-1] if spans else None  # type: Span

        # start profiling if profile_context is set
        if self.profile_status is None:
            self.profile_status = profile.profile_task_execution_service.add_profiling(self,
                                                                                       self.segment.segment_id,
                                                                                       op)

        if parent is not None and parent.kind.is_entry and inherit == parent.component:
            # Span's operation name could be override, recheck here
            # if the op name now is being profiling, start profile it here
            self.profiling_recheck(parent, op)

            span = parent
            span.op = op

        else:
            span = self.new_span(parent, EntrySpan, op=op)

            if carrier is not None and carrier.is_valid:  # TODO: should this be done irrespective of inheritance?
                span.extract(carrier=carrier)

        return span

    def new_exit_span(self, op: str, peer: str, component: Component = None, inherit: Component = None) -> Span:
        span = self.ignore_check(op, Kind.Exit)
        if span is not None:
            return span

        spans = _spans()
        parent = spans[-1] if spans else None  # type: Span

        if parent is not None and parent.kind.is_exit and component == parent.inherit:
            span = parent
            span.op = op
            span.peer = peer
            span.component = component

        else:
            span = self.new_span(parent, ExitSpan, op=op, peer=peer, component=component)

        if inherit:
            span.inherit = inherit

        return span

    def profiling_recheck(self, span: Span, op_name: str):
        # only check first span, e.g, first opname is correct.
        if span.sid != 0:
            return
        profile.profile_task_execution_service.profiling_recheck(self, self.segment.segment_id, op_name)

    def start(self, span: Span):
        self._nspans += 1
        spans = _spans_dup()
        if span not in spans:
            spans.append(span)

    def stop(self, span: Span) -> bool:
        spans = _spans_dup()
        span.finish(self.segment)

        try:
            spans.remove(span)
        except Exception:
            pass

        self._nspans -= 1
        if self._nspans == 0:
            agent.archive(self.segment)
            return True

        return False

    def active_span(self):
        spans = _spans()
        if spans:
            return spans[len(spans) - 1]

        return None

    def get_correlation(self, key):
        if key in self._correlation:
            return self._correlation[key]
        return None

    def put_correlation(self, key, value):
        if key is None:
            return
        if value is None:
            self._correlation.pop(key, value)
            return
        if len(value) > config.correlation_value_max_length:
            return
        if len(self._correlation) > config.correlation_element_max_number:
            return

        self._correlation[key] = value

    def capture(self):
        spans = _spans()
        if len(spans) == 0:
            return None

        return Snapshot(
            segment_id=str(self.segment.segment_id),
            span_id=self.active_span().sid,
            trace_id=self.segment.related_traces[0],
            endpoint=spans[0].op,
            correlation=self._correlation,
        )

    def continued(self, snapshot: 'Snapshot'):
        if snapshot is None:
            return None
        if not snapshot.is_from_current(self) and snapshot.is_valid():
            ref = SegmentRef.build_ref(snapshot)
            span = self.active_span()
            span.refs.append(ref)
            self.segment.relate(ID(ref.trace_id))
            self._correlation.update(snapshot.correlation)


class NoopContext(SpanContext):
    def __init__(self):
        super().__init__()

    def new_local_span(self, op: str) -> Span:
        return NoopSpan(self)

    def new_entry_span(self, op: str, carrier: 'Carrier' = None, inherit: Component = None) -> Span:
        return NoopSpan(self)

    def new_exit_span(self, op: str, peer: str, component: Component = None, inherit: Component = None) -> Span:
        return NoopSpan(self)

    def stop(self, span: Span) -> bool:
        spans = _spans_dup()

        try:
            spans.remove(span)
        except Exception:
            pass

        self._nspans -= 1

        return self._nspans == 0

    def capture(self):
        return Snapshot(
            segment_id=None,
            span_id=-1,
            trace_id=None,
            endpoint=None,
            correlation=self._correlation,
        )

    def continued(self, snapshot: 'Snapshot'):
        self._correlation.update(snapshot.correlation)


def get_context() -> SpanContext:
    spans = _spans()

    if spans:
        return spans[len(spans) - 1].context

    return SpanContext()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import io
import logging
import traceback


class SWFormatter(logging.Formatter):
    """ A slightly modified formatter that allows traceback depth """

    def __init__(self, fmt, tb_limit):
        logging.Formatter.__init__(self, fmt)
        self.tb_limit = tb_limit

    def formatException(self, ei):  # noqa
        sio = io.StringIO()
        tb = ei[2]
        traceback.print_exception(ei[0], ei[1], tb, self.tb_limit, sio)
        s = sio.getvalue()
        sio.close()
        if s[-1:] == '\n':
            s = s[:-1]
        return s



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ..common import Common_pb2 as common_dot_Common__pb2
from ..profile import Profile_pb2 as profile_dot_Profile__pb2


class ProfileTaskStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.getProfileTaskCommands = channel.unary_unary(
                '/skywalking.v3.ProfileTask/getProfileTaskCommands',
                request_serializer=profile_dot_Profile__pb2.ProfileTaskCommandQuery.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )
        self.collectSnapshot = channel.stream_unary(
                '/skywalking.v3.ProfileTask/collectSnapshot',
                request_serializer=profile_dot_Profile__pb2.ThreadSnapshot.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )
        self.reportTaskFinish = channel.unary_unary(
                '/skywalking.v3.ProfileTask/reportTaskFinish',
                request_serializer=profile_dot_Profile__pb2.ProfileTaskFinishReport.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )


class ProfileTaskServicer(object):
    """Missing associated documentation comment in .proto file."""

    def getProfileTaskCommands(self, request, context):
        """query all sniffer need to execute profile task commands
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def collectSnapshot(self, request_iterator, context):
        """collect dumped thread snapshot
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def reportTaskFinish(self, request, context):
        """report profiling task finished
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ProfileTaskServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'getProfileTaskCommands': grpc.unary_unary_rpc_method_handler(
                    servicer.getProfileTaskCommands,
                    request_deserializer=profile_dot_Profile__pb2.ProfileTaskCommandQuery.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
            'collectSnapshot': grpc.stream_unary_rpc_method_handler(
                    servicer.collectSnapshot,
                    request_deserializer=profile_dot_Profile__pb2.ThreadSnapshot.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
            'reportTaskFinish': grpc.unary_unary_rpc_method_handler(
                    servicer.reportTaskFinish,
                    request_deserializer=profile_dot_Profile__pb2.ProfileTaskFinishReport.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'skywalking.v3.ProfileTask', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ProfileTask(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def getProfileTaskCommands(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/skywalking.v3.ProfileTask/getProfileTaskCommands',
            profile_dot_Profile__pb2.ProfileTaskCommandQuery.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def collectSnapshot(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/skywalking.v3.ProfileTask/collectSnapshot',
            profile_dot_Profile__pb2.ThreadSnapshot.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def reportTaskFinish(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/skywalking.v3.ProfileTask/reportTaskFinish',
            profile_dot_Profile__pb2.ProfileTaskFinishReport.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: service-mesh-probe/service-mesh.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%service-mesh-probe/service-mesh.proto\x12\rskywalking.v3\x1a\x13\x63ommon/Common.proto\"\x9d\x03\n\x11ServiceMeshMetric\x12\x11\n\tstartTime\x18\x01 \x01(\x03\x12\x0f\n\x07\x65ndTime\x18\x02 \x01(\x03\x12\x19\n\x11sourceServiceName\x18\x03 \x01(\t\x12\x1d\n\x15sourceServiceInstance\x18\x04 \x01(\t\x12\x17\n\x0f\x64\x65stServiceName\x18\x05 \x01(\t\x12\x1b\n\x13\x64\x65stServiceInstance\x18\x06 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x07 \x01(\t\x12\x0f\n\x07latency\x18\x08 \x01(\x05\x12\x14\n\x0cresponseCode\x18\t \x01(\x05\x12\x0e\n\x06status\x18\n \x01(\x08\x12)\n\x08protocol\x18\x0b \x01(\x0e\x32\x17.skywalking.v3.Protocol\x12/\n\x0b\x64\x65tectPoint\x18\x0c \x01(\x0e\x32\x1a.skywalking.v3.DetectPoint\x12\x0f\n\x07tlsMode\x18\r \x01(\t\x12\x19\n\x11internalErrorCode\x18\x0e \x01(\t\x12#\n\x03tcp\x18\x0f \x01(\x0b\x32\x16.skywalking.v3.TCPInfo\"3\n\x07TCPInfo\x12\x15\n\rreceivedBytes\x18\x01 \x01(\x03\x12\x11\n\tsentBytes\x18\x02 \x01(\x03\"\x15\n\x13MeshProbeDownstream*\'\n\x08Protocol\x12\x08\n\x04HTTP\x10\x00\x12\x08\n\x04gRPC\x10\x01\x12\x07\n\x03TCP\x10\x02\x32o\n\x18ServiceMeshMetricService\x12S\n\x07\x63ollect\x12 .skywalking.v3.ServiceMeshMetric\x1a\".skywalking.v3.MeshProbeDownstream\"\x00(\x01\x42\x8d\x01\n0org.apache.skywalking.apm.network.servicemesh.v3P\x01Z7skywalking.apache.org/repo/goapi/collect/servicemesh/v3\xaa\x02\x1dSkyWalking.NetworkProtocol.V3b\x06proto3')

_PROTOCOL = DESCRIPTOR.enum_types_by_name['Protocol']
Protocol = enum_type_wrapper.EnumTypeWrapper(_PROTOCOL)
HTTP = 0
gRPC = 1
TCP = 2


_SERVICEMESHMETRIC = DESCRIPTOR.message_types_by_name['ServiceMeshMetric']
_TCPINFO = DESCRIPTOR.message_types_by_name['TCPInfo']
_MESHPROBEDOWNSTREAM = DESCRIPTOR.message_types_by_name['MeshProbeDownstream']
ServiceMeshMetric = _reflection.GeneratedProtocolMessageType('ServiceMeshMetric', (_message.Message,), {
  'DESCRIPTOR' : _SERVICEMESHMETRIC,
  '__module__' : 'service_mesh_probe.service_mesh_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.ServiceMeshMetric)
  })
_sym_db.RegisterMessage(ServiceMeshMetric)

TCPInfo = _reflection.GeneratedProtocolMessageType('TCPInfo', (_message.Message,), {
  'DESCRIPTOR' : _TCPINFO,
  '__module__' : 'service_mesh_probe.service_mesh_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.TCPInfo)
  })
_sym_db.RegisterMessage(TCPInfo)

MeshProbeDownstream = _reflection.GeneratedProtocolMessageType('MeshProbeDownstream', (_message.Message,), {
  'DESCRIPTOR' : _MESHPROBEDOWNSTREAM,
  '__module__' : 'service_mesh_probe.service_mesh_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.MeshProbeDownstream)
  })
_sym_db.RegisterMessage(MeshProbeDownstream)

_SERVICEMESHMETRICSERVICE = DESCRIPTOR.services_by_name['ServiceMeshMetricService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n0org.apache.skywalking.apm.network.servicemesh.v3P\001Z7skywalking.apache.org/repo/goapi/collect/servicemesh/v3\252\002\035SkyWalking.NetworkProtocol.V3'
  _PROTOCOL._serialized_start=569
  _PROTOCOL._serialized_end=608
  _SERVICEMESHMETRIC._serialized_start=78
  _SERVICEMESHMETRIC._serialized_end=491
  _TCPINFO._serialized_start=493
  _TCPINFO._serialized_end=544
  _MESHPROBEDOWNSTREAM._serialized_start=546
  _MESHPROBEDOWNSTREAM._serialized_end=567
  _SERVICEMESHMETRICSERVICE._serialized_start=610
  _SERVICEMESHMETRICSERVICE._serialized_end=721
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.context import get_context
from skywalking.trace.tags import TagDbType, TagDbInstance, TagDbStatement, TagDbSqlParameters

link_vector = ['https://pymysql.readthedocs.io/en/latest/']
support_matrix = {
    'pymysql': {
        '>=3.6': ['1.0']
    }
}
note = """"""


def install():
    from pymysql.cursors import Cursor

    _execute = Cursor.execute

    def _sw_execute(this: Cursor, query, args=None):
        peer = f'{this.connection.host}:{this.connection.port}'

        context = get_context()
        with context.new_exit_span(op='Mysql/PyMsql/execute', peer=peer, component=Component.PyMysql) as span:
            span.layer = Layer.Database
            res = _execute(this, query, args)

            span.tag(TagDbType('mysql'))
            span.tag(TagDbInstance((this.connection.db or b'').decode('utf-8')))
            span.tag(TagDbStatement(query))

            if config.sql_parameters_length and args:
                parameter = ','.join([str(arg) for arg in args])
                max_len = config.sql_parameters_length
                parameter = f'{parameter[0:max_len]}...' if len(parameter) > max_len else parameter
                span.tag(TagDbSqlParameters(f'[{parameter}]'))

            return res

    Cursor.execute = _sw_execute



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

class Tag:
    overridable = True

    def __init__(self, val):
        self.val = val


class TagHttpMethod(Tag):
    key = 'http.method'


class TagHttpURL(Tag):
    key = 'http.url'


class TagHttpStatusCode(Tag):
    key = 'http.status.code'


class TagHttpStatusMsg(Tag):
    key = 'http.status.msg'


class TagHttpParams(Tag):
    key = 'http.params'


class TagDbType(Tag):
    key = 'db.type'


class TagDbInstance(Tag):
    key = 'db.instance'


class TagDbStatement(Tag):
    key = 'db.statement'


class TagDbSqlParameters(Tag):
    key = 'db.sql.parameters'
    overridable = False


class TagMqBroker(Tag):
    key = 'mq.broker'


class TagMqTopic(Tag):
    key = 'mq.topic'


class TagMqQueue(Tag):
    key = 'mq.queue'


class TagCeleryParameters(Tag):
    key = 'celery.parameters'



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import atexit
import os
from queue import Queue, Full
from threading import Thread, Event
from typing import TYPE_CHECKING

from skywalking.protocol.logging.Logging_pb2 import LogData

from skywalking import config, plugins
from skywalking import loggings
from skywalking import profile
from skywalking.agent.protocol import Protocol
from skywalking.command import command_service
from skywalking.loggings import logger
from skywalking.profile.profile_task import ProfileTask
from skywalking.profile.snapshot import TracingThreadSnapshot

if TYPE_CHECKING:
    from skywalking.trace.context import Segment

__started = False
__protocol = Protocol()  # type: Protocol
__heartbeat_thread = __report_thread = __log_report_thread = __query_profile_thread = __command_dispatch_thread \
    = __send_profile_thread = __queue = __log_queue = __snapshot_queue = __finished = None


def __heartbeat():
    wait = base = 30

    while not __finished.is_set():
        try:
            __protocol.heartbeat()
            wait = base  # reset to base wait time on success
        except Exception as exc:
            logger.error(str(exc))
            wait = min(60, wait * 2 or 1)  # double wait time with each consecutive error up to a maximum

        __finished.wait(wait)


def __report():
    wait = base = 0

    while not __finished.is_set():
        try:
            __protocol.report(__queue)  # is blocking actually, blocks for max config.QUEUE_TIMEOUT seconds
            wait = base
        except Exception as exc:
            logger.error(str(exc))
            wait = min(60, wait * 2 or 1)

        __finished.wait(wait)


def __report_log():
    wait = base = 0

    while not __finished.is_set():
        try:
            __protocol.report_log(__log_queue)
            wait = base
        except Exception as exc:
            logger.error(str(exc))
            wait = min(60, wait * 2 or 1)

        __finished.wait(wait)


def __send_profile_snapshot():
    wait = base = 0.5

    while not __finished.is_set():
        try:
            __protocol.send_snapshot(__snapshot_queue)
            wait = base
        except Exception as exc:
            logger.error(str(exc))
            wait = min(60, wait * 2 or 1)

        __finished.wait(wait)


def __query_profile_command():
    wait = base = config.get_profile_task_interval

    while not __finished.is_set():
        try:
            __protocol.query_profile_commands()
            wait = base
        except Exception as exc:
            logger.error(str(exc))
            wait = min(60, wait * 2 or 1)

        __finished.wait(wait)


def __command_dispatch():
    # command dispatch will stuck when there are no commands
    command_service.dispatch()


def __init_threading():
    global __heartbeat_thread, __report_thread, __log_report_thread, __query_profile_thread, \
        __command_dispatch_thread, __send_profile_thread, __queue, __log_queue, __snapshot_queue, __finished

    __queue = Queue(maxsize=config.max_buffer_size)
    __finished = Event()
    __heartbeat_thread = Thread(name='HeartbeatThread', target=__heartbeat, daemon=True)
    __report_thread = Thread(name='ReportThread', target=__report, daemon=True)
    __command_dispatch_thread = Thread(name='CommandDispatchThread', target=__command_dispatch, daemon=True)

    __heartbeat_thread.start()
    __report_thread.start()
    __command_dispatch_thread.start()

    if config.log_reporter_active:
        __log_queue = Queue(maxsize=config.log_reporter_max_buffer_size)
        __log_report_thread = Thread(name='LogReportThread', target=__report_log, daemon=True)
        __log_report_thread.start()

    if config.profile_active:
        __snapshot_queue = Queue(maxsize=config.profile_snapshot_transport_buffer_size)

        __query_profile_thread = Thread(name='QueryProfileCommandThread', target=__query_profile_command, daemon=True)
        __query_profile_thread.start()

        __send_profile_thread = Thread(name='SendProfileSnapShotThread', target=__send_profile_snapshot, daemon=True)
        __send_profile_thread.start()


def __init():
    global __protocol
    if config.protocol == 'grpc':
        from skywalking.agent.protocol.grpc import GrpcProtocol
        __protocol = GrpcProtocol()
    elif config.protocol == 'http':
        from skywalking.agent.protocol.http import HttpProtocol
        __protocol = HttpProtocol()
    elif config.protocol == 'kafka':
        from skywalking.agent.protocol.kafka import KafkaProtocol
        __protocol = KafkaProtocol()

    plugins.install()
    if config.log_reporter_active:  # todo - Add support for printing traceID/ context in logs
        from skywalking import log
        log.install()

    __init_threading()


def __fini():
    __protocol.report(__queue, False)
    __queue.join()

    if config.log_reporter_active:
        __protocol.report_log(__log_queue, False)
        __log_queue.join()

    if config.profile_active:
        __protocol.send_snapshot(__snapshot_queue, False)
        __snapshot_queue.join()

    __finished.set()


def __fork_before():
    if config.protocol != 'http':
        logger.warning(f'fork() not currently supported with {config.protocol} protocol')

    # TODO: handle __queue and __finished correctly (locks, mutexes, etc...), need to lock before fork and unlock after
    # if possible, or ensure they are not locked in threads (end threads and restart after fork?)

    __protocol.fork_before()


def __fork_after_in_parent():
    __protocol.fork_after_in_parent()


def __fork_after_in_child():
    __protocol.fork_after_in_child()
    __init_threading()


def start():
    global __started
    if __started:
        return
    __started = True

    flag = False
    try:
        from gevent import monkey
        flag = monkey.is_module_patched('socket')
    except ModuleNotFoundError:
        logger.debug("it was found that no gevent was used, if you don't use, please ignore.")
    if flag:
        import grpc.experimental.gevent as grpc_gevent
        grpc_gevent.init_gevent()

    loggings.init()
    config.finalize()
    profile.init()

    __init()

    atexit.register(__fini)

    if (hasattr(os, 'register_at_fork')):
        os.register_at_fork(before=__fork_before, after_in_parent=__fork_after_in_parent,
                            after_in_child=__fork_after_in_child)


def stop():
    atexit.unregister(__fini)
    __fini()


def started():
    return __started


def isfull():
    return __queue.full()


def archive(segment: 'Segment'):
    try:  # unlike checking __queue.full() then inserting, this is atomic
        __queue.put(segment, block=False)
    except Full:
        logger.warning('the queue is full, the segment will be abandoned')


def archive_log(log_data: 'LogData'):
    try:
        __log_queue.put(log_data, block=False)
    except Full:
        logger.warning('the queue is full, the log will be abandoned')


def add_profiling_snapshot(snapshot: TracingThreadSnapshot):
    try:
        __snapshot_queue.put(snapshot)
    except Full:
        logger.warning('the snapshot queue is full, the snapshot will be abandoned')


def notify_profile_finish(task: ProfileTask):
    try:
        __protocol.notify_profile_task_finish(task)
    except Exception as e:
        logger.error(f'notify profile task finish to backend fail. {str(e)}')



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import queue
from collections import deque

from skywalking.protocol.common.Common_pb2 import Commands, Command

from skywalking.command.base_command import BaseCommand
from skywalking.command.executors import noop_command_executor_instance
from skywalking.command.executors.profile_task_command_executor import ProfileTaskCommandExecutor
from skywalking.command.profile_task_command import ProfileTaskCommand
from skywalking.loggings import logger


class CommandService:

    def __init__(self):
        self._commands = queue.Queue()  # type: queue.Queue
        # don't execute same command twice
        self._command_serial_number_cache = CommandSerialNumberCache()

    def dispatch(self):
        while True:
            # block until a command is available
            command = self._commands.get()  # type: BaseCommand
            if not self.__is_command_executed(command):
                command_executor_service.execute(command)
                self._command_serial_number_cache.add(command.serial_number)

    def __is_command_executed(self, command: BaseCommand):
        return self._command_serial_number_cache.contains(command.serial_number)

    def receive_command(self, commands: Commands):
        for command in commands.commands:
            try:
                base_command = CommandDeserializer.deserialize(command)
                logger.debug('received command [{%s} {%s}]', base_command.command, base_command.serial_number)

                if self.__is_command_executed(base_command):
                    logger.warning('command[{%s}] is executed, ignored.', base_command.command)
                    continue

                try:
                    self._commands.put(base_command)
                except queue.Full:
                    logger.warning('command[{%s}, {%s}] cannot add to command list. because the command list is full.',
                                   base_command.command, base_command.serial_number)
            except UnsupportedCommandException as e:
                logger.warning('received unsupported command[{%s}].', e.command.command)


class CommandSerialNumberCache:

    def __init__(self, maxlen=64):
        self.queue = deque(maxlen=maxlen)

    def add(self, number: str):
        # Once a bounded length deque is full, when new items are added,
        # a corresponding number of items are discarded from the opposite end.
        self.queue.append(number)

    def contains(self, number: str) -> bool:
        try:
            _ = self.queue.index(number)
            return True
        except ValueError:
            return False


class CommandExecutorService:
    """
    route commands to appropriate executor
    """

    def __init__(self):
        self.__command_executor_map = {ProfileTaskCommand.NAME: ProfileTaskCommandExecutor()}

    def execute(self, command: BaseCommand):
        self.__executor_for_command(command).execute(command)

    def __executor_for_command(self, command: BaseCommand):
        executor = self.__command_executor_map.get(command.command)
        if not executor:
            return noop_command_executor_instance
        return executor


class CommandDeserializer:

    @staticmethod
    def deserialize(command: Command) -> BaseCommand:
        command_name = command.command

        if ProfileTaskCommand.NAME == command_name:
            return ProfileTaskCommand.deserialize(command)
        else:
            raise UnsupportedCommandException(command)


class UnsupportedCommandException(Exception):

    def __init__(self, command):
        self.command = command


# init
command_executor_service = CommandExecutorService()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class ProfileConstants:
    # Monitor duration must greater than 1 minutes
    TASK_DURATION_MIN_MINUTE = 1
    # The duration of the monitoring task cannot be greater than 15 minutes
    TASK_DURATION_MAX_MINUTE = 15
    # Dump period must be greater than or equals 10 milliseconds
    TASK_DUMP_PERIOD_MIN_MILLIS = 10
    # Max sampling count must less than 10
    TASK_MAX_SAMPLING_COUNT = 10



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: logging/Logging.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15logging/Logging.proto\x12\rskywalking.v3\x1a\x13\x63ommon/Common.proto\"\xea\x01\n\x07LogData\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x0f\n\x07service\x18\x02 \x01(\t\x12\x17\n\x0fserviceInstance\x18\x03 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x04 \x01(\t\x12(\n\x04\x62ody\x18\x05 \x01(\x0b\x32\x1a.skywalking.v3.LogDataBody\x12\x31\n\x0ctraceContext\x18\x06 \x01(\x0b\x32\x1b.skywalking.v3.TraceContext\x12$\n\x04tags\x18\x07 \x01(\x0b\x32\x16.skywalking.v3.LogTags\x12\r\n\x05layer\x18\x08 \x01(\t\"\x9e\x01\n\x0bLogDataBody\x12\x0c\n\x04type\x18\x01 \x01(\t\x12&\n\x04text\x18\x02 \x01(\x0b\x32\x16.skywalking.v3.TextLogH\x00\x12&\n\x04json\x18\x03 \x01(\x0b\x32\x16.skywalking.v3.JSONLogH\x00\x12&\n\x04yaml\x18\x04 \x01(\x0b\x32\x16.skywalking.v3.YAMLLogH\x00\x42\t\n\x07\x63ontent\"\x17\n\x07TextLog\x12\x0c\n\x04text\x18\x01 \x01(\t\"\x17\n\x07JSONLog\x12\x0c\n\x04json\x18\x01 \x01(\t\"\x17\n\x07YAMLLog\x12\x0c\n\x04yaml\x18\x01 \x01(\t\"G\n\x0cTraceContext\x12\x0f\n\x07traceId\x18\x01 \x01(\t\x12\x16\n\x0etraceSegmentId\x18\x02 \x01(\t\x12\x0e\n\x06spanId\x18\x03 \x01(\x05\":\n\x07LogTags\x12/\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32!.skywalking.v3.KeyStringValuePair2R\n\x10LogReportService\x12>\n\x07\x63ollect\x12\x16.skywalking.v3.LogData\x1a\x17.skywalking.v3.Commands\"\x00(\x01\x42\x85\x01\n,org.apache.skywalking.apm.network.logging.v3P\x01Z3skywalking.apache.org/repo/goapi/collect/logging/v3\xaa\x02\x1dSkyWalking.NetworkProtocol.V3b\x06proto3')



_LOGDATA = DESCRIPTOR.message_types_by_name['LogData']
_LOGDATABODY = DESCRIPTOR.message_types_by_name['LogDataBody']
_TEXTLOG = DESCRIPTOR.message_types_by_name['TextLog']
_JSONLOG = DESCRIPTOR.message_types_by_name['JSONLog']
_YAMLLOG = DESCRIPTOR.message_types_by_name['YAMLLog']
_TRACECONTEXT = DESCRIPTOR.message_types_by_name['TraceContext']
_LOGTAGS = DESCRIPTOR.message_types_by_name['LogTags']
LogData = _reflection.GeneratedProtocolMessageType('LogData', (_message.Message,), {
  'DESCRIPTOR' : _LOGDATA,
  '__module__' : 'logging.Logging_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.LogData)
  })
_sym_db.RegisterMessage(LogData)

LogDataBody = _reflection.GeneratedProtocolMessageType('LogDataBody', (_message.Message,), {
  'DESCRIPTOR' : _LOGDATABODY,
  '__module__' : 'logging.Logging_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.LogDataBody)
  })
_sym_db.RegisterMessage(LogDataBody)

TextLog = _reflection.GeneratedProtocolMessageType('TextLog', (_message.Message,), {
  'DESCRIPTOR' : _TEXTLOG,
  '__module__' : 'logging.Logging_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.TextLog)
  })
_sym_db.RegisterMessage(TextLog)

JSONLog = _reflection.GeneratedProtocolMessageType('JSONLog', (_message.Message,), {
  'DESCRIPTOR' : _JSONLOG,
  '__module__' : 'logging.Logging_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.JSONLog)
  })
_sym_db.RegisterMessage(JSONLog)

YAMLLog = _reflection.GeneratedProtocolMessageType('YAMLLog', (_message.Message,), {
  'DESCRIPTOR' : _YAMLLOG,
  '__module__' : 'logging.Logging_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.YAMLLog)
  })
_sym_db.RegisterMessage(YAMLLog)

TraceContext = _reflection.GeneratedProtocolMessageType('TraceContext', (_message.Message,), {
  'DESCRIPTOR' : _TRACECONTEXT,
  '__module__' : 'logging.Logging_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.TraceContext)
  })
_sym_db.RegisterMessage(TraceContext)

LogTags = _reflection.GeneratedProtocolMessageType('LogTags', (_message.Message,), {
  'DESCRIPTOR' : _LOGTAGS,
  '__module__' : 'logging.Logging_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.LogTags)
  })
_sym_db.RegisterMessage(LogTags)

_LOGREPORTSERVICE = DESCRIPTOR.services_by_name['LogReportService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n,org.apache.skywalking.apm.network.logging.v3P\001Z3skywalking.apache.org/repo/goapi/collect/logging/v3\252\002\035SkyWalking.NetworkProtocol.V3'
  _LOGDATA._serialized_start=62
  _LOGDATA._serialized_end=296
  _LOGDATABODY._serialized_start=299
  _LOGDATABODY._serialized_end=457
  _TEXTLOG._serialized_start=459
  _TEXTLOG._serialized_end=482
  _JSONLOG._serialized_start=484
  _JSONLOG._serialized_end=507
  _YAMLLOG._serialized_start=509
  _YAMLLOG._serialized_end=532
  _TRACECONTEXT._serialized_start=534
  _TRACECONTEXT._serialized_end=605
  _LOGTAGS._serialized_start=607
  _LOGTAGS._serialized_end=665
  _LOGREPORTSERVICE._serialized_start=667
  _LOGREPORTSERVICE._serialized_end=749
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component
from skywalking.trace.context import get_context
from skywalking.trace.tags import TagDbType, TagDbInstance, TagDbStatement

link_vector = ['https://github.com/andymccurdy/redis-py/']
support_matrix = {
    'redis': {
        '>=3.6': ['3.5']  # "4.0" next, incompatible to current instrumentation
    }
}
note = """"""


def install():
    from redis.connection import Connection

    _send_command = Connection.send_command

    def _sw_send_command(this: Connection, *args, **kwargs):
        peer = f'{this.host}:{this.port}'
        op = args[0]
        context = get_context()
        with context.new_exit_span(op=f'Redis/{op}' or '/', peer=peer, component=Component.Redis) as span:
            span.layer = Layer.Cache

            res = _send_command(this, *args, **kwargs)
            span.tag(TagDbType('Redis'))
            span.tag(TagDbInstance(this.db))
            span.tag(TagDbStatement(op))

            return res

    Connection.send_command = _sw_send_command



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List

from skywalking import config
from skywalking.utils.lang import b64encode, b64decode


class CarrierItem(object):
    def __init__(self, key: str = '', val: str = ''):
        self.key = f'{config.agent_namespace}-{key}' if config.agent_namespace else key  # type: str
        self.val = val  # type: str

    @property
    def key(self):
        return self.__key

    @key.setter
    def key(self, key: str):
        self.__key = key

    @property
    def val(self):
        return self.__val

    @val.setter
    def val(self, val: str):
        self.__val = val


class Carrier(CarrierItem):
    def __init__(self, trace_id: str = '', segment_id: str = '', span_id: str = '', service: str = '',
                 service_instance: str = '', endpoint: str = '', client_address: str = '',
                 correlation: dict = None):  # pyre-ignore
        super(Carrier, self).__init__(key='sw8')
        self.__val = None
        self.trace_id = trace_id  # type: str
        self.segment_id = segment_id  # type: str
        self.span_id = span_id  # type: str
        self.service = service  # type: str
        self.service_instance = service_instance  # type: str
        self.endpoint = endpoint  # type: str
        self.client_address = client_address  # type: str
        self.correlation_carrier = SW8CorrelationCarrier()
        self.items = [self.correlation_carrier, self]  # type: List[CarrierItem]
        self.__iter_index = 0  # type: int
        if correlation is not None:
            self.correlation_carrier.correlation = correlation

    @property
    def val(self) -> str:
        return '-'.join([
            '1',
            b64encode(self.trace_id),
            b64encode(self.segment_id),
            self.span_id,
            b64encode(self.service),
            b64encode(self.service_instance),
            b64encode(self.endpoint),
            b64encode(self.client_address),
        ])

    @val.setter
    def val(self, val: str):
        self.__val = val
        if not val:
            return
        parts = val.split('-')
        if len(parts) != 8:
            return
        self.trace_id = b64decode(parts[1])
        self.segment_id = b64decode(parts[2])
        self.span_id = parts[3]
        self.service = b64decode(parts[4])
        self.service_instance = b64decode(parts[5])
        self.endpoint = b64decode(parts[6])
        self.client_address = b64decode(parts[7])

    @property
    def is_valid(self):
        # type: () -> bool
        return len(self.trace_id) > 0 and \
               len(self.segment_id) > 0 and \
               len(self.service) > 0 and \
               len(self.service_instance) > 0 and \
               len(self.endpoint) > 0 and \
               len(self.client_address) > 0 and \
               self.span_id.isnumeric()

    @property
    def is_suppressed(self):  # if is invalid from previous set, ignored or suppressed status propagation downstream
        return self.__val and not self.is_valid

    def __iter__(self):
        self.__iter_index = 0
        return self

    def __next__(self):
        if self.__iter_index >= len(self.items):
            raise StopIteration
        n = self.items[self.__iter_index]
        self.__iter_index += 1
        return n


class SW8CorrelationCarrier(CarrierItem):
    def __init__(self):
        super(SW8CorrelationCarrier, self).__init__(key='sw8-correlation')
        self.correlation = {}  # type: dict

    @property
    def val(self) -> str:
        if self.correlation is None or len(self.correlation) == 0:
            return ''

        return ','.join([
            f'{b64encode(k)}:{b64encode(v)}'
            for k, v in self.correlation.items()
        ])

    @val.setter
    def val(self, val: str):
        self.__val = val
        if not val:
            return
        for per in val.split(','):
            if len(self.correlation) > config.correlation_element_max_number:
                break
            parts = per.split(':')
            if len(parts) != 2:
                continue
            self.correlation[b64decode(parts[0])] = b64decode(parts[1])



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
from abc import ABC
from collections import defaultdict
from typing import List, Union, DefaultDict
from typing import TYPE_CHECKING

from skywalking import Kind, Layer, Log, Component, LogItem, config
from skywalking.trace import ID
from skywalking.trace.carrier import Carrier
from skywalking.trace.segment import SegmentRef, Segment
from skywalking.trace.tags import Tag
from skywalking.utils.lang import tostring

if TYPE_CHECKING:
    from skywalking.trace.context import SpanContext


@tostring
class Span(ABC):
    def __init__(
            self,
            context: 'SpanContext',
            sid: int = -1,
            pid: int = -1,
            op: str = None,
            peer: str = None,
            kind: Kind = None,
            component: Component = None,
            layer: Layer = None,
    ):
        self._depth = 0
        self.context = context  # type: SpanContext
        self.sid = sid  # type: int
        self.pid = pid  # type: int
        self.op = op  # type: str
        self.peer = peer  # type: str
        self.kind = kind  # type: Kind
        self.component = component or Component.Unknown  # type: Component
        self.layer = layer or Layer.Unknown  # type: Layer
        self.inherit = Component.Unknown  # type: Component

        self.tags = defaultdict(list)  # type: DefaultDict[str, Union[Tag, List[Tag]]]
        self.logs = []  # type: List[Log]
        self.refs = []  # type: List[SegmentRef]
        self.start_time = 0  # type: int
        self.end_time = 0  # type: int
        self.error_occurred = False  # type: bool

    def start(self):
        self._depth += 1
        if self._depth != 1:
            return

        self.start_time = int(time.time() * 1000)
        self.context.start(self)

    def stop(self):
        self._depth -= 1
        if self._depth:
            return False

        return self.context.stop(self)

    def finish(self, segment: 'Segment') -> bool:
        self.end_time = int(time.time() * 1000)
        segment.archive(self)
        return True

    def raised(self) -> 'Span':
        from skywalking.utils.filter import sw_traceback
        self.error_occurred = True
        self.logs = [Log(items=[
            LogItem(key='Traceback', val=sw_traceback()),
        ])]
        return self

    def log(self, ex: Exception) -> 'Span':
        self.error_occurred = True
        self.logs.append(Log(items=[LogItem(key='Traceback', val=str(ex))]))
        return self

    def tag(self, tag: Tag) -> 'Span':
        if tag.overridable:
            self.tags[tag.key] = tag
        else:
            self.tags[tag.key].append(tag)

        return self

    def iter_tags(self):
        for tag in self.tags.values():
            if isinstance(tag, Tag):
                yield tag
            else:
                yield from tag

    def inject(self) -> 'Carrier':
        raise RuntimeWarning(
            'can only inject context carrier into ExitSpan, this may be a potential bug in the agent, '
            'please report this in https://github.com/apache/skywalking/issues if you encounter this. '
        )

    def extract(self, carrier: 'Carrier') -> 'Span':
        if carrier is None:
            return self

        self.context.segment.relate(ID(carrier.trace_id))
        self.context._correlation = carrier.correlation_carrier.correlation

        if not carrier.is_valid:
            return self

        ref = SegmentRef(carrier=carrier)

        if ref not in self.refs:
            self.refs.append(ref)

        return self

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if isinstance(exc_val, Exception):
            self.raised()
        self.stop()
        if exc_tb is not None:
            return False
        return True


@tostring
class EntrySpan(Span):
    def __init__(
            self,
            context: 'SpanContext',
            sid: int = -1,
            pid: int = -1,
            op: str = None,
            peer: str = None,
            component: 'Component' = None,
            layer: 'Layer' = None,
    ):
        Span.__init__(
            self,
            context,
            sid,
            pid,
            op,
            peer,
            Kind.Entry,
            component,
            layer,
        )
        self._max_depth = 0

    def start(self):
        Span.start(self)
        self._max_depth = self._depth
        self.component = 0
        self.layer = Layer.Unknown
        self.logs = []
        self.tags = defaultdict(list)


@tostring
class ExitSpan(Span):
    def __init__(
            self,
            context: 'SpanContext',
            sid: int = -1,
            pid: int = -1,
            op: str = None,
            peer: str = None,
            component: 'Component' = None,
            layer: 'Layer' = None,
    ):
        Span.__init__(
            self,
            context,
            sid,
            pid,
            op,
            peer,
            Kind.Exit,
            component,
            layer,
        )

    def inject(self) -> 'Carrier':
        return Carrier(
            trace_id=str(self.context.segment.related_traces[0]),
            segment_id=str(self.context.segment.segment_id),
            span_id=str(self.sid),
            service=config.service_name,
            service_instance=config.service_instance,
            endpoint=self.op,
            client_address=self.peer,
            correlation=self.context._correlation,
        )


@tostring
class NoopSpan(Span):
    def __init__(self, context: 'SpanContext' = None):
        Span.__init__(self, context=context, op='', kind=Kind.Local)

    def extract(self, carrier: 'Carrier'):
        return

    def inject(self) -> 'Carrier':
        return Carrier()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: service-mesh-probe/service-mesh-compat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..service_mesh_probe import service_mesh_pb2 as service__mesh__probe_dot_service__mesh__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,service-mesh-probe/service-mesh-compat.proto\x1a%service-mesh-probe/service-mesh.proto2o\n\x18ServiceMeshMetricService\x12S\n\x07\x63ollect\x12 .skywalking.v3.ServiceMeshMetric\x1a\".skywalking.v3.MeshProbeDownstream\"\x00(\x01\x42\x9e\x01\n7org.apache.skywalking.apm.network.servicemesh.v3.compatP\x01Z7skywalking.apache.org/repo/goapi/collect/servicemesh/v3\xb8\x01\x01\xaa\x02$SkyWalking.NetworkProtocol.V3.Compatb\x06proto3')



_SERVICEMESHMETRICSERVICE = DESCRIPTOR.services_by_name['ServiceMeshMetricService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n7org.apache.skywalking.apm.network.servicemesh.v3.compatP\001Z7skywalking.apache.org/repo/goapi/collect/servicemesh/v3\270\001\001\252\002$SkyWalking.NetworkProtocol.V3.Compat'
  _SERVICEMESHMETRICSERVICE._serialized_start=87
  _SERVICEMESHMETRICSERVICE._serialized_end=198
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging

from skywalking import Layer, Component, config
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context, NoopContext
from skywalking.trace.span import NoopSpan
from skywalking.trace.tags import TagHttpMethod, TagHttpURL, TagHttpStatusCode, TagHttpParams

logger = logging.getLogger(__name__)

# version_rule = {
#     "name": "sanic",
#     "rules": [">=20.3.0 <21.0.0"]
# }

link_vector = ['https://sanic.readthedocs.io/en/latest']
support_matrix = {
    'sanic': {
        '>=3.10': [],  # not supporting any version yet
        '>=3.7': ['20.12'],  # 21.9 Future LTS - Not supported by SW yet
        '>=3.6': ['20.12']  # 20.12 last LTS for python 3.6
    }  # TODO: add instrumentation for 21.9 (method signature change) remove - write_callback, stream_callback
}
note = """"""


def install():
    from sanic import Sanic, handlers, response
    # TODO: format_http1_response is removed from response in later versions.

    _format_http1_response = response.format_http1_response
    _handle_request = Sanic.handle_request
    _handlers_ErrorHandler_response = handlers.ErrorHandler.response # noqa

    def _sw_format_http1_response(status: int, headers, body=b''):
        if status is not None:
            entry_span = get_context().active_span()
            if entry_span is not None and type(entry_span) is not NoopSpan:
                if status >= 400:
                    entry_span.error_occurred = True
                entry_span.tag(TagHttpStatusCode(status))

        return _format_http1_response(status, headers, body)

    def _sw_handlers_ErrorHandler_response(self: handlers.ErrorHandler, req, e): # noqa
        if e is not None:
            entry_span = get_context().active_span()
            if entry_span is not None and type(entry_span) is not NoopSpan:
                entry_span.raised()

        return _handlers_ErrorHandler_response(self, req, e)

    response.format_http1_response = _sw_format_http1_response
    Sanic.handle_request = _gen_sw_handle_request(_handle_request)
    handlers.ErrorHandler.response = _sw_handlers_ErrorHandler_response


def _gen_sw_handle_request(_handle_request):
    from inspect import isawaitable

    def params_tostring(params):
        return '\n'.join([f"{k}=[{','.join(params.getlist(k))}]" for k, _ in params.items()])

    async def _sw_handle_request(self, request, write_callback, stream_callback):
        req = request
        carrier = Carrier()
        method = req.method

        for item in carrier:
            if item.key.capitalize() in req.headers:
                item.val = req.headers[item.key.capitalize()]

        span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
            else get_context().new_entry_span(op=req.path, carrier=carrier)

        with span:
            span.layer = Layer.Http
            span.component = Component.Sanic
            span.peer = f'{req.remote_addr or req.ip}:{req.port}'
            span.tag(TagHttpMethod(method))
            span.tag(TagHttpURL(req.url.split('?')[0]))
            if config.sanic_collect_http_params and req.args:
                span.tag(TagHttpParams(params_tostring(req.args)[0:config.http_params_length_threshold]))
            resp = _handle_request(self, request, write_callback, stream_callback)
            if isawaitable(resp):
                result = await resp

        return result

    return _sw_handle_request



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: profile/Profile.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15profile/Profile.proto\x12\rskywalking.v3\x1a\x13\x63ommon/Common.proto\"\\\n\x17ProfileTaskCommandQuery\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x17\n\x0fserviceInstance\x18\x02 \x01(\t\x12\x17\n\x0flastCommandTime\x18\x03 \x01(\x03\"\x83\x01\n\x0eThreadSnapshot\x12\x0e\n\x06taskId\x18\x01 \x01(\t\x12\x16\n\x0etraceSegmentId\x18\x02 \x01(\t\x12\x0c\n\x04time\x18\x03 \x01(\x03\x12\x10\n\x08sequence\x18\x04 \x01(\x05\x12)\n\x05stack\x18\x05 \x01(\x0b\x32\x1a.skywalking.v3.ThreadStack\"%\n\x0bThreadStack\x12\x16\n\x0e\x63odeSignatures\x18\x01 \x03(\t\"S\n\x17ProfileTaskFinishReport\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x17\n\x0fserviceInstance\x18\x02 \x01(\t\x12\x0e\n\x06taskId\x18\x03 \x01(\t2\x90\x02\n\x0bProfileTask\x12[\n\x16getProfileTaskCommands\x12&.skywalking.v3.ProfileTaskCommandQuery\x1a\x17.skywalking.v3.Commands\"\x00\x12M\n\x0f\x63ollectSnapshot\x12\x1d.skywalking.v3.ThreadSnapshot\x1a\x17.skywalking.v3.Commands\"\x00(\x01\x12U\n\x10reportTaskFinish\x12&.skywalking.v3.ProfileTaskFinishReport\x1a\x17.skywalking.v3.Commands\"\x00\x42\x97\x01\n5org.apache.skywalking.apm.network.language.profile.v3P\x01Z<skywalking.apache.org/repo/goapi/collect/language/profile/v3\xaa\x02\x1dSkyWalking.NetworkProtocol.V3b\x06proto3')



_PROFILETASKCOMMANDQUERY = DESCRIPTOR.message_types_by_name['ProfileTaskCommandQuery']
_THREADSNAPSHOT = DESCRIPTOR.message_types_by_name['ThreadSnapshot']
_THREADSTACK = DESCRIPTOR.message_types_by_name['ThreadStack']
_PROFILETASKFINISHREPORT = DESCRIPTOR.message_types_by_name['ProfileTaskFinishReport']
ProfileTaskCommandQuery = _reflection.GeneratedProtocolMessageType('ProfileTaskCommandQuery', (_message.Message,), {
  'DESCRIPTOR' : _PROFILETASKCOMMANDQUERY,
  '__module__' : 'profile.Profile_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.ProfileTaskCommandQuery)
  })
_sym_db.RegisterMessage(ProfileTaskCommandQuery)

ThreadSnapshot = _reflection.GeneratedProtocolMessageType('ThreadSnapshot', (_message.Message,), {
  'DESCRIPTOR' : _THREADSNAPSHOT,
  '__module__' : 'profile.Profile_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.ThreadSnapshot)
  })
_sym_db.RegisterMessage(ThreadSnapshot)

ThreadStack = _reflection.GeneratedProtocolMessageType('ThreadStack', (_message.Message,), {
  'DESCRIPTOR' : _THREADSTACK,
  '__module__' : 'profile.Profile_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.ThreadStack)
  })
_sym_db.RegisterMessage(ThreadStack)

ProfileTaskFinishReport = _reflection.GeneratedProtocolMessageType('ProfileTaskFinishReport', (_message.Message,), {
  'DESCRIPTOR' : _PROFILETASKFINISHREPORT,
  '__module__' : 'profile.Profile_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.ProfileTaskFinishReport)
  })
_sym_db.RegisterMessage(ProfileTaskFinishReport)

_PROFILETASK = DESCRIPTOR.services_by_name['ProfileTask']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n5org.apache.skywalking.apm.network.language.profile.v3P\001Z<skywalking.apache.org/repo/goapi/collect/language/profile/v3\252\002\035SkyWalking.NetworkProtocol.V3'
  _PROFILETASKCOMMANDQUERY._serialized_start=61
  _PROFILETASKCOMMANDQUERY._serialized_end=153
  _THREADSNAPSHOT._serialized_start=156
  _THREADSNAPSHOT._serialized_end=287
  _THREADSTACK._serialized_start=289
  _THREADSTACK._serialized_end=326
  _PROFILETASKFINISHREPORT._serialized_start=328
  _PROFILETASKFINISHREPORT._serialized_end=411
  _PROFILETASK._serialized_start=414
  _PROFILETASK._serialized_end=686
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: profile/ProfileCompat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2
from ..profile import Profile_pb2 as profile_dot_Profile__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bprofile/ProfileCompat.proto\x1a\x13\x63ommon/Common.proto\x1a\x15profile/Profile.proto2\x90\x02\n\x0bProfileTask\x12[\n\x16getProfileTaskCommands\x12&.skywalking.v3.ProfileTaskCommandQuery\x1a\x17.skywalking.v3.Commands\"\x00\x12M\n\x0f\x63ollectSnapshot\x12\x1d.skywalking.v3.ThreadSnapshot\x1a\x17.skywalking.v3.Commands\"\x00(\x01\x12U\n\x10reportTaskFinish\x12&.skywalking.v3.ProfileTaskFinishReport\x1a\x17.skywalking.v3.Commands\"\x00\x42\xa8\x01\n<org.apache.skywalking.apm.network.language.profile.v3.compatP\x01Z<skywalking.apache.org/repo/goapi/collect/language/profile/v3\xb8\x01\x01\xaa\x02$SkyWalking.NetworkProtocol.V3.Compatb\x06proto3')



_PROFILETASK = DESCRIPTOR.services_by_name['ProfileTask']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n<org.apache.skywalking.apm.network.language.profile.v3.compatP\001Z<skywalking.apache.org/repo/goapi/collect/language/profile/v3\270\001\001\252\002$SkyWalking.NetworkProtocol.V3.Compat'
  _PROFILETASK._serialized_start=76
  _PROFILETASK._serialized_end=348
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component
from skywalking import config
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context
from skywalking.trace.tags import TagMqBroker, TagMqTopic

link_vector = ['https://kafka-python.readthedocs.io']

support_matrix = {
    'kafka-python': {
        '>=3.6': ['2.0']
    }
}
note = """"""


def install():
    from kafka import KafkaProducer
    from kafka import KafkaConsumer

    _send = KafkaProducer.send
    __poll_once = KafkaConsumer._poll_once
    KafkaProducer.send = _sw_send_func(_send)
    KafkaConsumer._poll_once = _sw__poll_once_func(__poll_once)


def _sw__poll_once_func(__poll_once):
    def _sw__poll_once(this, timeout_ms, max_records, update_offsets=True):
        res = __poll_once(this, timeout_ms, max_records, update_offsets=update_offsets)
        if res:
            brokers = ';'.join(this.config['bootstrap_servers'])
            context = get_context()
            topics = ';'.join(this._subscription.subscription
                              or [t.topic for t in this._subscription._user_assignment])

            with context.new_entry_span(
                    op=f"Kafka/{topics}/Consumer/{this.config['group_id'] or ''}") as span:
                for consumer_records in res.values():
                    for record in consumer_records:
                        carrier = Carrier()
                        for item in carrier:
                            for header in record.headers:
                                if item.key == header[0]:
                                    item.val = str(header[1])

                        span.extract(carrier)
                    span.tag(TagMqBroker(brokers))
                    span.tag(TagMqTopic(topics))
                    span.layer = Layer.MQ
                    span.component = Component.KafkaConsumer

        return res

    return _sw__poll_once


def _sw_send_func(_send):
    def _sw_send(this, topic, value=None, key=None, headers=None, partition=None, timestamp_ms=None):
        # ignore trace & log reporter - skywalking self request
        if config.protocol == 'kafka' and config.kafka_topic_segment == topic \
                or config.kafka_topic_log == topic \
                or config.kafka_topic_management == topic:
            return _send(this, topic, value=value, key=key, headers=headers, partition=partition,
                         timestamp_ms=timestamp_ms)

        peer = ';'.join(this.config['bootstrap_servers'])
        context = get_context()
        with context.new_exit_span(op=f'Kafka/{topic}/Producer' or '/', peer=peer,
                                   component=Component.KafkaProducer) as span:
            carrier = span.inject()
            span.layer = Layer.MQ

            if headers is None:
                headers = []
            for item in carrier:
                headers.append((item.key, item.val.encode('utf-8')))

            res = _send(this, topic, value=value, key=key, headers=headers, partition=partition,
                        timestamp_ms=timestamp_ms)
            span.tag(TagMqBroker(peer))
            span.tag(TagMqTopic(topic))

            return res

    return _sw_send



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
from queue import Queue, Empty
from time import time

from skywalking import config
from skywalking.agent import Protocol
from skywalking.client.kafka import KafkaServiceManagementClient, KafkaTraceSegmentReportService, \
    KafkaLogDataReportService
from skywalking.loggings import logger, getLogger, logger_debug_enabled
from skywalking.protocol.common.Common_pb2 import KeyStringValuePair
from skywalking.protocol.language_agent.Tracing_pb2 import SegmentObject, SpanObject, Log, SegmentReference
from skywalking.protocol.logging.Logging_pb2 import LogData
from skywalking.trace.segment import Segment

# avoid too many kafka logs
logger_kafka = getLogger('kafka')
logger_kafka.setLevel(max(logging.WARN, logger.level))


class KafkaProtocol(Protocol):
    def __init__(self):
        self.service_management = KafkaServiceManagementClient()
        self.traces_reporter = KafkaTraceSegmentReportService()
        self.log_reporter = KafkaLogDataReportService()

    def heartbeat(self):
        self.service_management.send_heart_beat()

    def report(self, queue: Queue, block: bool = True):
        start = None

        def generator():
            nonlocal start

            while True:
                try:
                    timeout = config.QUEUE_TIMEOUT  # type: int
                    if not start:  # make sure first time through queue is always checked
                        start = time()
                    else:
                        timeout -= int(time() - start)
                        if timeout <= 0:  # this is to make sure we exit eventually instead of being fed continuously
                            return
                    segment = queue.get(block=block, timeout=timeout)  # type: Segment
                except Empty:
                    return

                queue.task_done()
                if logger_debug_enabled:
                    logger.debug('reporting segment %s', segment)

                s = SegmentObject(
                    traceId=str(segment.related_traces[0]),
                    traceSegmentId=str(segment.segment_id),
                    service=config.service_name,
                    serviceInstance=config.service_instance,
                    spans=[SpanObject(
                        spanId=span.sid,
                        parentSpanId=span.pid,
                        startTime=span.start_time,
                        endTime=span.end_time,
                        operationName=span.op,
                        peer=span.peer,
                        spanType=span.kind.name,
                        spanLayer=span.layer.name,
                        componentId=span.component.value,
                        isError=span.error_occurred,
                        logs=[Log(
                            time=int(log.timestamp * 1000),
                            data=[KeyStringValuePair(key=item.key, value=item.val) for item in log.items],
                        ) for log in span.logs],
                        tags=[KeyStringValuePair(
                            key=tag.key,
                            value=str(tag.val),
                        ) for tag in span.iter_tags()],
                        refs=[SegmentReference(
                            refType=0 if ref.ref_type == 'CrossProcess' else 1,
                            traceId=ref.trace_id,
                            parentTraceSegmentId=ref.segment_id,
                            parentSpanId=ref.span_id,
                            parentService=ref.service,
                            parentServiceInstance=ref.service_instance,
                            parentEndpoint=ref.endpoint,
                            networkAddressUsedAtPeer=ref.client_address,
                        ) for ref in span.refs if ref.trace_id],
                    ) for span in segment.spans],
                )

                yield s

        self.traces_reporter.report(generator())

    def report_log(self, queue: Queue, block: bool = True):
        start = None

        def generator():
            nonlocal start

            while True:
                try:
                    timeout = config.QUEUE_TIMEOUT  # type: int
                    if not start:  # make sure first time through queue is always checked
                        start = time()
                    else:
                        timeout -= int(time() - start)
                        if timeout <= 0:  # this is to make sure we exit eventually instead of being fed continuously
                            return
                    log_data = queue.get(block=block, timeout=timeout)  # type: LogData
                except Empty:
                    return
                queue.task_done()

                if logger_debug_enabled:
                    logger.debug('Reporting Log')

                yield log_data

        self.log_reporter.report(generator=generator())



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context
from skywalking.trace.tags import TagMqBroker, TagCeleryParameters

link_vector = ['https://docs.celeryproject.org']
# TODO: Celery is missing plugin test
support_matrix = {
    'celery': {
        '>=3.6': ['5.1']
    }
}
note = """The celery server running with "celery -A ..." should be run with the HTTP protocol
as it uses multiprocessing by default which is not compatible with the gRPC protocol implementation
in SkyWalking currently. Celery clients can use whatever protocol they want."""


def install():
    from urllib.parse import urlparse
    from celery import Celery

    def send_task(self, name, args=None, kwargs=None, **options):
        # NOTE: Lines commented out below left for documentation purposes if sometime in the future exchange / queue
        # names are wanted. Currently these do not match between producer and consumer so would need some work.

        broker_url = self.conf['broker_url']
        # exchange = options['exchange']
        # queue = options['routing_key']
        # op = 'celery/{}/{}/{}'.format(exchange or '', queue or '', name)
        op = f'celery/{name}'

        if broker_url:
            url = urlparse(broker_url)
            peer = f'{url.hostname}:{url.port}'
        else:
            peer = '???'

        with get_context().new_exit_span(op=op, peer=peer, component=Component.Celery) as span:
            span.layer = Layer.MQ

            span.tag(TagMqBroker(broker_url))
            # span.tag(TagMqTopic(exchange))
            # span.tag(TagMqQueue(queue))

            if config.celery_parameters_length:
                params = f'*{args}, **{kwargs}'[:config.celery_parameters_length]
                span.tag(TagCeleryParameters(params))

            options = {**options}
            headers = options.get('headers')
            headers = {**headers} if headers else {}
            options['headers'] = headers

            for item in span.inject():
                headers[item.key] = item.val

            return _send_task(self, name, args, kwargs, **options)

    _send_task = Celery.send_task
    Celery.send_task = send_task

    def task_from_fun(self, _fun, name=None, **options):
        def fun(*args, **kwargs):
            req = task.request_stack.top
            # di = req.get('delivery_info')
            # exchange = di and di.get('exchange')
            # queue = di and di.get('routing_key')
            # op = 'celery/{}/{}/{}'.format(exchange or '', queue or '', name)
            op = f'celery/{name}'
            carrier = Carrier()

            for item in carrier:
                val = req.get(item.key)

                if val:
                    item.val = val

            context = get_context()
            origin = req.get('origin')

            if origin:
                span = context.new_entry_span(op=op, carrier=carrier)
                span.peer = origin.split('@', 1)[-1]
            else:
                span = context.new_local_span(op=op)

            with span:
                span.layer = Layer.MQ
                span.component = Component.Celery

                span.tag(TagMqBroker(task.app.conf['broker_url']))
                # span.tag(TagMqTopic(exchange))
                # span.tag(TagMqQueue(queue))

                if config.celery_parameters_length:
                    params = f'*{args}, **{kwargs}'[:config.celery_parameters_length]
                    span.tag(TagCeleryParameters(params))

                return _fun(*args, **kwargs)

        name = name or self.gen_task_name(_fun.__name__, _fun.__module__)
        task = _task_from_fun(self, fun, name, **options)

        return task

    _task_from_fun = Celery._task_from_fun
    Celery._task_from_fun = task_from_fun



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ..browser import BrowserPerf_pb2 as browser_dot_BrowserPerf__pb2
from ..common import Common_pb2 as common_dot_Common__pb2


class BrowserPerfServiceStub(object):
    """Collect performance raw data from browser.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.collectPerfData = channel.unary_unary(
                '/BrowserPerfService/collectPerfData',
                request_serializer=browser_dot_BrowserPerf__pb2.BrowserPerfData.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )
        self.collectErrorLogs = channel.stream_unary(
                '/BrowserPerfService/collectErrorLogs',
                request_serializer=browser_dot_BrowserPerf__pb2.BrowserErrorLog.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )


class BrowserPerfServiceServicer(object):
    """Collect performance raw data from browser.
    """

    def collectPerfData(self, request, context):
        """report once per page
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def collectErrorLogs(self, request_iterator, context):
        """report one or more error logs for pages, could report multiple times.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_BrowserPerfServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'collectPerfData': grpc.unary_unary_rpc_method_handler(
                    servicer.collectPerfData,
                    request_deserializer=browser_dot_BrowserPerf__pb2.BrowserPerfData.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
            'collectErrorLogs': grpc.stream_unary_rpc_method_handler(
                    servicer.collectErrorLogs,
                    request_deserializer=browser_dot_BrowserPerf__pb2.BrowserErrorLog.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'BrowserPerfService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class BrowserPerfService(object):
    """Collect performance raw data from browser.
    """

    @staticmethod
    def collectPerfData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/BrowserPerfService/collectPerfData',
            browser_dot_BrowserPerf__pb2.BrowserPerfData.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def collectErrorLogs(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/BrowserPerfService/collectErrorLogs',
            browser_dot_BrowserPerf__pb2.BrowserErrorLog.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking.protocol.common.Common_pb2 import Command

from skywalking.command.base_command import BaseCommand
from skywalking.utils.lang import tostring


@tostring
class ProfileTaskCommand(BaseCommand):
    NAME = 'ProfileTaskQuery'

    def __init__(self,
                 serial_number: str = '',
                 task_id: str = '',
                 endpoint_name: str = '',
                 duration: int = -1,
                 min_duration_threshold: int = -1,
                 dump_period: int = -1,
                 max_sampling_count: int = -1,
                 start_time: int = -1,
                 create_time: int = -1):

        BaseCommand.__init__(self, self.NAME, serial_number)

        self.task_id = task_id  # type: str
        self.endpoint_name = endpoint_name  # type: str
        self.duration = duration  # type: int
        self.min_duration_threshold = min_duration_threshold  # type: int
        self.dump_period = dump_period  # type: int
        self.max_sampling_count = max_sampling_count  # type: int
        self.start_time = start_time  # type: int
        self.create_time = create_time  # type: int

    @staticmethod
    def deserialize(command: Command):
        serial_number = None
        task_id = None
        endpoint_name = None
        duration = None
        min_duration_threshold = None
        dump_period = None
        max_sampling_count = None
        start_time = None
        create_time = None

        for pair in command.args:
            if pair.key == 'SerialNumber':
                serial_number = pair.value
            elif pair.key == 'EndpointName':
                endpoint_name = pair.value
            elif pair.key == 'TaskId':
                task_id = pair.value
            elif pair.key == 'Duration':
                duration = pair.value
            elif pair.key == 'MinDurationThreshold':
                min_duration_threshold = pair.value
            elif pair.key == 'DumpPeriod':
                dump_period = pair.value
            elif pair.key == 'MaxSamplingCount':
                max_sampling_count = pair.value
            elif pair.key == 'StartTime':
                start_time = pair.value
            elif pair.key == 'CreateTime':
                create_time = pair.value

        return ProfileTaskCommand(serial_number=serial_number, task_id=task_id,
                                  endpoint_name=endpoint_name, duration=duration,
                                  min_duration_threshold=min_duration_threshold, dump_period=dump_period,
                                  max_sampling_count=max_sampling_count, start_time=start_time,
                                  create_time=create_time)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: management/ManagementCompat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2
from ..management import Management_pb2 as management_dot_Management__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!management/ManagementCompat.proto\x1a\x13\x63ommon/Common.proto\x1a\x1bmanagement/Management.proto2\xb5\x01\n\x11ManagementService\x12X\n\x18reportInstanceProperties\x12!.skywalking.v3.InstanceProperties\x1a\x17.skywalking.v3.Commands\"\x00\x12\x46\n\tkeepAlive\x12\x1e.skywalking.v3.InstancePingPkg\x1a\x17.skywalking.v3.Commands\"\x00\x42\x9c\x01\n6org.apache.skywalking.apm.network.management.v3.compatP\x01Z6skywalking.apache.org/repo/goapi/collect/management/v3\xb8\x01\x01\xaa\x02$SkyWalking.NetworkProtocol.V3.Compatb\x06proto3')



_MANAGEMENTSERVICE = DESCRIPTOR.services_by_name['ManagementService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n6org.apache.skywalking.apm.network.management.v3.compatP\001Z6skywalking.apache.org/repo/goapi/collect/management/v3\270\001\001\252\002$SkyWalking.NetworkProtocol.V3.Compat'
  _MANAGEMENTSERVICE._serialized_start=88
  _MANAGEMENTSERVICE._serialized_end=269
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ..common import Common_pb2 as common_dot_Common__pb2
from ..event import Event_pb2 as event_dot_Event__pb2


class EventServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.collect = channel.stream_unary(
                '/skywalking.v3.EventService/collect',
                request_serializer=event_dot_Event__pb2.Event.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )


class EventServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def collect(self, request_iterator, context):
        """When reporting an event, you typically call the collect function twice, one for starting of the event and the other one for ending of the event, with the same UUID.
        There are also cases where you have both start time and end time already, for example, when exporting events from a 3rd-party system,
        the start time and end time are already known so that you can call the collect function only once.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_EventServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'collect': grpc.stream_unary_rpc_method_handler(
                    servicer.collect,
                    request_deserializer=event_dot_Event__pb2.Event.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'skywalking.v3.EventService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class EventService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def collect(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/skywalking.v3.EventService/collect',
            event_dot_Event__pb2.Event.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: language-agent/TracingCompat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2
from ..language_agent import Tracing_pb2 as language__agent_dot_Tracing__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"language-agent/TracingCompat.proto\x1a\x13\x63ommon/Common.proto\x1a\x1clanguage-agent/Tracing.proto2\xaf\x01\n\x19TraceSegmentReportService\x12\x44\n\x07\x63ollect\x12\x1c.skywalking.v3.SegmentObject\x1a\x17.skywalking.v3.Commands\"\x00(\x01\x12L\n\rcollectInSync\x12 .skywalking.v3.SegmentCollection\x1a\x17.skywalking.v3.Commands\"\x00\x42\xa4\x01\n:org.apache.skywalking.apm.network.language.agent.v3.compatP\x01Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\xb8\x01\x01\xaa\x02$SkyWalking.NetworkProtocol.V3.Compatb\x06proto3')



_TRACESEGMENTREPORTSERVICE = DESCRIPTOR.services_by_name['TraceSegmentReportService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n:org.apache.skywalking.apm.network.language.agent.v3.compatP\001Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\270\001\001\252\002$SkyWalking.NetworkProtocol.V3.Compat'
  _TRACESEGMENTREPORTSERVICE._serialized_start=90
  _TRACESEGMENTREPORTSERVICE._serialized_end=265
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: language-agent/CLRMetric.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1elanguage-agent/CLRMetric.proto\x12\rskywalking.v3\x1a\x13\x63ommon/Common.proto\"j\n\x13\x43LRMetricCollection\x12)\n\x07metrics\x18\x01 \x03(\x0b\x32\x18.skywalking.v3.CLRMetric\x12\x0f\n\x07service\x18\x02 \x01(\t\x12\x17\n\x0fserviceInstance\x18\x03 \x01(\t\"\x86\x01\n\tCLRMetric\x12\x0c\n\x04time\x18\x01 \x01(\x03\x12\x1f\n\x03\x63pu\x18\x02 \x01(\x0b\x32\x12.skywalking.v3.CPU\x12 \n\x02gc\x18\x03 \x01(\x0b\x32\x14.skywalking.v3.ClrGC\x12(\n\x06thread\x18\x04 \x01(\x0b\x32\x18.skywalking.v3.ClrThread\"i\n\x05\x43lrGC\x12\x18\n\x10Gen0CollectCount\x18\x01 \x01(\x03\x12\x18\n\x10Gen1CollectCount\x18\x02 \x01(\x03\x12\x18\n\x10Gen2CollectCount\x18\x03 \x01(\x03\x12\x12\n\nHeapMemory\x18\x04 \x01(\x03\"\x8f\x01\n\tClrThread\x12&\n\x1e\x41vailableCompletionPortThreads\x18\x01 \x01(\x05\x12\x1e\n\x16\x41vailableWorkerThreads\x18\x02 \x01(\x05\x12 \n\x18MaxCompletionPortThreads\x18\x03 \x01(\x05\x12\x18\n\x10MaxWorkerThreads\x18\x04 \x01(\x05\x32\x62\n\x16\x43LRMetricReportService\x12H\n\x07\x63ollect\x12\".skywalking.v3.CLRMetricCollection\x1a\x17.skywalking.v3.Commands\"\x00\x42\x93\x01\n3org.apache.skywalking.apm.network.language.agent.v3P\x01Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\xaa\x02\x1dSkyWalking.NetworkProtocol.V3b\x06proto3')



_CLRMETRICCOLLECTION = DESCRIPTOR.message_types_by_name['CLRMetricCollection']
_CLRMETRIC = DESCRIPTOR.message_types_by_name['CLRMetric']
_CLRGC = DESCRIPTOR.message_types_by_name['ClrGC']
_CLRTHREAD = DESCRIPTOR.message_types_by_name['ClrThread']
CLRMetricCollection = _reflection.GeneratedProtocolMessageType('CLRMetricCollection', (_message.Message,), {
  'DESCRIPTOR' : _CLRMETRICCOLLECTION,
  '__module__' : 'language_agent.CLRMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.CLRMetricCollection)
  })
_sym_db.RegisterMessage(CLRMetricCollection)

CLRMetric = _reflection.GeneratedProtocolMessageType('CLRMetric', (_message.Message,), {
  'DESCRIPTOR' : _CLRMETRIC,
  '__module__' : 'language_agent.CLRMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.CLRMetric)
  })
_sym_db.RegisterMessage(CLRMetric)

ClrGC = _reflection.GeneratedProtocolMessageType('ClrGC', (_message.Message,), {
  'DESCRIPTOR' : _CLRGC,
  '__module__' : 'language_agent.CLRMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.ClrGC)
  })
_sym_db.RegisterMessage(ClrGC)

ClrThread = _reflection.GeneratedProtocolMessageType('ClrThread', (_message.Message,), {
  'DESCRIPTOR' : _CLRTHREAD,
  '__module__' : 'language_agent.CLRMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.ClrThread)
  })
_sym_db.RegisterMessage(ClrThread)

_CLRMETRICREPORTSERVICE = DESCRIPTOR.services_by_name['CLRMetricReportService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n3org.apache.skywalking.apm.network.language.agent.v3P\001Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\252\002\035SkyWalking.NetworkProtocol.V3'
  _CLRMETRICCOLLECTION._serialized_start=70
  _CLRMETRICCOLLECTION._serialized_end=176
  _CLRMETRIC._serialized_start=179
  _CLRMETRIC._serialized_end=313
  _CLRGC._serialized_start=315
  _CLRGC._serialized_end=420
  _CLRTHREAD._serialized_start=423
  _CLRTHREAD._serialized_end=566
  _CLRMETRICREPORTSERVICE._serialized_start=568
  _CLRMETRICREPORTSERVICE._serialized_end=666
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import inspect

from skywalking import Layer, Component, config
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context, NoopContext
from skywalking.trace.span import NoopSpan
from skywalking.trace.tags import TagHttpMethod, TagHttpURL, TagHttpStatusCode

link_vector = ['https://docs.python.org/3/library/http.server.html',
               'https://werkzeug.palletsprojects.com/']
support_matrix = {
    'http_server': {
        '>=3.6': ['*']
    },
    'werkzeug': {
        '>=3.6': ['1.0.1', '2.0']
    }
}
note = """"""


def install():
    from http.server import BaseHTTPRequestHandler

    _handle = BaseHTTPRequestHandler.handle

    def _sw_handle(handler: BaseHTTPRequestHandler):
        clazz = handler.__class__
        if 'werkzeug.serving.WSGIRequestHandler' == '.'.join([clazz.__module__, clazz.__name__]):
            wrap_werkzeug_request_handler(handler)
        else:
            wrap_default_request_handler(handler)
        _handle(handler)

    BaseHTTPRequestHandler.handle = _sw_handle

    def _sw_send_response_only(self, code, *args, **kwargs):
        self._status_code = code

        return _send_response_only(self, code, *args, **kwargs)

    _send_response_only = BaseHTTPRequestHandler.send_response_only
    BaseHTTPRequestHandler.send_response_only = _sw_send_response_only


def wrap_werkzeug_request_handler(handler):
    """
    Wrap run_wsgi of werkzeug.serving.WSGIRequestHandler to add skywalking instrument code.
    """
    _run_wsgi = handler.run_wsgi

    def _wrap_run_wsgi():
        carrier = Carrier()
        method = handler.command

        for item in carrier:
            item.val = handler.headers[item.key.capitalize()]
        path = handler.path or '/'

        span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
            else get_context().new_entry_span(op=path.split('?')[0], carrier=carrier)

        with span:
            url = f"http://{handler.headers['Host']}{path}" if 'Host' in handler.headers else path
            span.layer = Layer.Http
            span.component = Component.General
            client_address = handler.client_address
            span.peer = f'{client_address[0]}:{client_address[1]}'
            span.tag(TagHttpMethod(method))
            span.tag(TagHttpURL(url))

            try:
                return _run_wsgi()
            finally:
                status_code = int(getattr(handler, '_status_code', -1))
                if status_code > -1:
                    span.tag(TagHttpStatusCode(status_code))
                    if status_code >= 400:
                        span.error_occurred = True

    handler.run_wsgi = _wrap_run_wsgi

    def _sw_send_response(self, code, *args, **kwargs):
        self._status_code = code

        return _send_response(self, code, *args, **kwargs)

    WSGIRequestHandler = handler.__class__  # noqa

    if not getattr(WSGIRequestHandler, '_sw_wrapped', False):
        _send_response = WSGIRequestHandler.send_response
        WSGIRequestHandler.send_response = _sw_send_response
        WSGIRequestHandler._sw_wrapped = True


def wrap_default_request_handler(handler):
    http_methods = ('GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH')
    for method in http_methods:
        _wrap_do_method(handler, method)


def _wrap_do_method(handler, method):
    if hasattr(handler, f'do_{method}') and inspect.ismethod(getattr(handler, f'do_{method}')):
        _do_method = getattr(handler, f'do_{method}')

        def _sw_do_method():
            carrier = Carrier()
            for item in carrier:
                item.val = handler.headers[item.key.capitalize()]
            path = handler.path or '/'

            span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
                else get_context().new_entry_span(op=path.split('?')[0], carrier=carrier)

            with span:
                url = f"http://{handler.headers['Host']}{path}" if 'Host' in handler.headers else path
                span.layer = Layer.Http
                span.component = Component.General
                client_address = handler.client_address
                span.peer = f'{client_address[0]}:{client_address[1]}'
                span.tag(TagHttpMethod(method))
                span.tag(TagHttpURL(url))

                try:
                    _do_method()
                finally:
                    status_code = int(getattr(handler, '_status_code', -1))
                    if status_code > -1:
                        span.tag(TagHttpStatusCode(status_code))
                        if status_code >= 400:
                            span.error_occurred = True

        setattr(handler, f'do_{method}', _sw_do_method)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json

import requests
from google.protobuf import json_format

from skywalking import config
from skywalking.client import ServiceManagementClient, TraceSegmentReportService, LogDataReportService
from skywalking.loggings import logger, logger_debug_enabled


class HttpServiceManagementClient(ServiceManagementClient):
    def __init__(self):
        proto = 'https://' if config.force_tls else 'http://'
        self.url_instance_props = f"{proto}{config.collector_address.rstrip('/')}/v3/management/reportProperties"
        self.url_heart_beat = f"{proto}{config.collector_address.rstrip('/')}/v3/management/keepAlive"
        self.session = requests.Session()

    def fork_after_in_child(self):
        self.session.close()
        self.session = requests.Session()

    def send_instance_props(self):
        res = self.session.post(self.url_instance_props, json={
            'service': config.service_name,
            'serviceInstance': config.service_instance,
            'properties': [{
                'language': 'Python',
            }]
        })
        if logger_debug_enabled:
            logger.debug('heartbeat response: %s', res)

    def send_heart_beat(self):
        if logger_debug_enabled:
            logger.debug(
                'service heart beats, [%s], [%s]',
                config.service_name,
                config.service_instance,
            )
        res = self.session.post(self.url_heart_beat, json={
            'service': config.service_name,
            'serviceInstance': config.service_instance,
        })
        if logger_debug_enabled:
            logger.debug('heartbeat response: %s', res)


class HttpTraceSegmentReportService(TraceSegmentReportService):
    def __init__(self):
        proto = 'https://' if config.force_tls else 'http://'
        self.url_report = f"{proto}{config.collector_address.rstrip('/')}/v3/segment"
        self.session = requests.Session()

    def fork_after_in_child(self):
        self.session.close()
        self.session = requests.Session()

    def report(self, generator):
        for segment in generator:
            res = self.session.post(self.url_report, json={
                'traceId': str(segment.related_traces[0]),
                'traceSegmentId': str(segment.segment_id),
                'service': config.service_name,
                'serviceInstance': config.service_instance,
                'spans': [{
                    'spanId': span.sid,
                    'parentSpanId': span.pid,
                    'startTime': span.start_time,
                    'endTime': span.end_time,
                    'operationName': span.op,
                    'peer': span.peer,
                    'spanType': span.kind.name,
                    'spanLayer': span.layer.name,
                    'componentId': span.component.value,
                    'isError': span.error_occurred,
                    'logs': [{
                        'time': int(log.timestamp * 1000),
                        'data': [{
                            'key': item.key,
                            'value': item.val,
                        } for item in log.items],
                    } for log in span.logs],
                    'tags': [{
                        'key': tag.key,
                        'value': tag.val,
                    } for tag in span.iter_tags()],
                    'refs': [{
                        'refType': 0,
                        'traceId': ref.trace_id,
                        'parentTraceSegmentId': ref.segment_id,
                        'parentSpanId': ref.span_id,
                        'parentService': ref.service,
                        'parentServiceInstance': ref.service_instance,
                        'parentEndpoint': ref.endpoint,
                        'networkAddressUsedAtPeer': ref.client_address,
                    } for ref in span.refs if ref.trace_id]
                } for span in segment.spans]
            })
            if logger_debug_enabled:
                logger.debug('report traces response: %s', res)


class HttpLogDataReportService(LogDataReportService):
    def __init__(self):
        proto = 'https://' if config.force_tls else 'http://'
        self.url_report = f"{proto}{config.collector_address.rstrip('/')}/v3/logs"
        self.session = requests.Session()

    def fork_after_in_child(self):
        self.session.close()
        self.session = requests.Session()

    def report(self, generator):
        log_batch = [json.loads(json_format.MessageToJson(log_data)) for log_data in generator]
        if log_batch:  # prevent empty batches
            res = self.session.post(self.url_report, json=log_batch)
            if logger_debug_enabled:
                logger.debug('report batch log response: %s', res)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.context import get_context
from skywalking.trace.tags import TagDbType, TagDbStatement

link_vector = ['https://github.com/elastic/elasticsearch-py']
support_matrix = {
    'elasticsearch': {
        '>=3.6': ['7.13', '7.14', '7.15'],
    }
}
note = """"""


def install():
    from elasticsearch import Transport
    _perform_request = Transport.perform_request

    def _sw_perform_request(this: Transport, method, url, headers=None, params=None, body=None):
        context = get_context()
        peer = ','.join([f"{host['host']}:{str(host['port'])}" for host in this.hosts])
        with context.new_exit_span(op=f'Elasticsearch/{method}{url}', peer=peer,
                                   component=Component.Elasticsearch) as span:
            span.layer = Layer.Database
            res = _perform_request(this, method, url, headers=headers, params=params, body=body)

            span.tag(TagDbType('Elasticsearch'))
            if config.elasticsearch_trace_dsl:
                span.tag(TagDbStatement('' if body is None else body))

            return res

    Transport.perform_request = _sw_perform_request



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: management/Management.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bmanagement/Management.proto\x12\rskywalking.v3\x1a\x13\x63ommon/Common.proto\"u\n\x12InstanceProperties\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x17\n\x0fserviceInstance\x18\x02 \x01(\t\x12\x35\n\nproperties\x18\x03 \x03(\x0b\x32!.skywalking.v3.KeyStringValuePair\";\n\x0fInstancePingPkg\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x17\n\x0fserviceInstance\x18\x02 \x01(\t2\xb5\x01\n\x11ManagementService\x12X\n\x18reportInstanceProperties\x12!.skywalking.v3.InstanceProperties\x1a\x17.skywalking.v3.Commands\"\x00\x12\x46\n\tkeepAlive\x12\x1e.skywalking.v3.InstancePingPkg\x1a\x17.skywalking.v3.Commands\"\x00\x42\x8b\x01\n/org.apache.skywalking.apm.network.management.v3P\x01Z6skywalking.apache.org/repo/goapi/collect/management/v3\xaa\x02\x1dSkyWalking.NetworkProtocol.V3b\x06proto3')



_INSTANCEPROPERTIES = DESCRIPTOR.message_types_by_name['InstanceProperties']
_INSTANCEPINGPKG = DESCRIPTOR.message_types_by_name['InstancePingPkg']
InstanceProperties = _reflection.GeneratedProtocolMessageType('InstanceProperties', (_message.Message,), {
  'DESCRIPTOR' : _INSTANCEPROPERTIES,
  '__module__' : 'management.Management_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.InstanceProperties)
  })
_sym_db.RegisterMessage(InstanceProperties)

InstancePingPkg = _reflection.GeneratedProtocolMessageType('InstancePingPkg', (_message.Message,), {
  'DESCRIPTOR' : _INSTANCEPINGPKG,
  '__module__' : 'management.Management_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.InstancePingPkg)
  })
_sym_db.RegisterMessage(InstancePingPkg)

_MANAGEMENTSERVICE = DESCRIPTOR.services_by_name['ManagementService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n/org.apache.skywalking.apm.network.management.v3P\001Z6skywalking.apache.org/repo/goapi/collect/management/v3\252\002\035SkyWalking.NetworkProtocol.V3'
  _INSTANCEPROPERTIES._serialized_start=67
  _INSTANCEPROPERTIES._serialized_end=184
  _INSTANCEPINGPKG._serialized_start=186
  _INSTANCEPINGPKG._serialized_end=245
  _MANAGEMENTSERVICE._serialized_start=248
  _MANAGEMENTSERVICE._serialized_end=429
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ..common import Common_pb2 as common_dot_Common__pb2
from ..management import Management_pb2 as management_dot_Management__pb2


class ManagementServiceStub(object):
    """Define the service reporting the extra information of the instance.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.reportInstanceProperties = channel.unary_unary(
                '/ManagementService/reportInstanceProperties',
                request_serializer=management_dot_Management__pb2.InstanceProperties.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )
        self.keepAlive = channel.unary_unary(
                '/ManagementService/keepAlive',
                request_serializer=management_dot_Management__pb2.InstancePingPkg.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )


class ManagementServiceServicer(object):
    """Define the service reporting the extra information of the instance.
    """

    def reportInstanceProperties(self, request, context):
        """Report custom properties of a service instance.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def keepAlive(self, request, context):
        """Keep the instance alive in the backend analysis.
        Only recommend to do separate keepAlive report when no trace and metrics needs to be reported.
        Otherwise, it is duplicated.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ManagementServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'reportInstanceProperties': grpc.unary_unary_rpc_method_handler(
                    servicer.reportInstanceProperties,
                    request_deserializer=management_dot_Management__pb2.InstanceProperties.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
            'keepAlive': grpc.unary_unary_rpc_method_handler(
                    servicer.keepAlive,
                    request_deserializer=management_dot_Management__pb2.InstancePingPkg.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ManagementService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ManagementService(object):
    """Define the service reporting the extra information of the instance.
    """

    @staticmethod
    def reportInstanceProperties(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ManagementService/reportInstanceProperties',
            management_dot_Management__pb2.InstanceProperties.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def keepAlive(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ManagementService/keepAlive',
            management_dot_Management__pb2.InstancePingPkg.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: language-agent/JVMMetric.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1elanguage-agent/JVMMetric.proto\x12\rskywalking.v3\x1a\x13\x63ommon/Common.proto\"j\n\x13JVMMetricCollection\x12)\n\x07metrics\x18\x01 \x03(\x0b\x32\x18.skywalking.v3.JVMMetric\x12\x0f\n\x07service\x18\x02 \x01(\t\x12\x17\n\x0fserviceInstance\x18\x03 \x01(\t\"\xfb\x01\n\tJVMMetric\x12\x0c\n\x04time\x18\x01 \x01(\x03\x12\x1f\n\x03\x63pu\x18\x02 \x01(\x0b\x32\x12.skywalking.v3.CPU\x12%\n\x06memory\x18\x03 \x03(\x0b\x32\x15.skywalking.v3.Memory\x12-\n\nmemoryPool\x18\x04 \x03(\x0b\x32\x19.skywalking.v3.MemoryPool\x12\x1d\n\x02gc\x18\x05 \x03(\x0b\x32\x11.skywalking.v3.GC\x12%\n\x06thread\x18\x06 \x01(\x0b\x32\x15.skywalking.v3.Thread\x12#\n\x05\x63lazz\x18\x07 \x01(\x0b\x32\x14.skywalking.v3.Class\"T\n\x06Memory\x12\x0e\n\x06isHeap\x18\x01 \x01(\x08\x12\x0c\n\x04init\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x12\x0c\n\x04used\x18\x04 \x01(\x03\x12\x11\n\tcommitted\x18\x05 \x01(\x03\"o\n\nMemoryPool\x12%\n\x04type\x18\x01 \x01(\x0e\x32\x17.skywalking.v3.PoolType\x12\x0c\n\x04init\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x12\x0c\n\x04used\x18\x04 \x01(\x03\x12\x11\n\tcommitted\x18\x05 \x01(\x03\"H\n\x02GC\x12%\n\x05phase\x18\x01 \x01(\x0e\x32\x16.skywalking.v3.GCPhase\x12\r\n\x05\x63ount\x18\x02 \x01(\x03\x12\x0c\n\x04time\x18\x03 \x01(\x03\"\xcd\x01\n\x06Thread\x12\x11\n\tliveCount\x18\x01 \x01(\x03\x12\x13\n\x0b\x64\x61\x65monCount\x18\x02 \x01(\x03\x12\x11\n\tpeakCount\x18\x03 \x01(\x03\x12 \n\x18runnableStateThreadCount\x18\x04 \x01(\x03\x12\x1f\n\x17\x62lockedStateThreadCount\x18\x05 \x01(\x03\x12\x1f\n\x17waitingStateThreadCount\x18\x06 \x01(\x03\x12$\n\x1ctimedWaitingStateThreadCount\x18\x07 \x01(\x03\"a\n\x05\x43lass\x12\x18\n\x10loadedClassCount\x18\x01 \x01(\x03\x12\x1f\n\x17totalUnloadedClassCount\x18\x02 \x01(\x03\x12\x1d\n\x15totalLoadedClassCount\x18\x03 \x01(\x03*\x80\x01\n\x08PoolType\x12\x14\n\x10\x43ODE_CACHE_USAGE\x10\x00\x12\x10\n\x0cNEWGEN_USAGE\x10\x01\x12\x10\n\x0cOLDGEN_USAGE\x10\x02\x12\x12\n\x0eSURVIVOR_USAGE\x10\x03\x12\x11\n\rPERMGEN_USAGE\x10\x04\x12\x13\n\x0fMETASPACE_USAGE\x10\x05*\'\n\x07GCPhase\x12\x07\n\x03NEW\x10\x00\x12\x07\n\x03OLD\x10\x01\x12\n\n\x06NORMAL\x10\x02\x32\x62\n\x16JVMMetricReportService\x12H\n\x07\x63ollect\x12\".skywalking.v3.JVMMetricCollection\x1a\x17.skywalking.v3.Commands\"\x00\x42\x93\x01\n3org.apache.skywalking.apm.network.language.agent.v3P\x01Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\xaa\x02\x1dSkyWalking.NetworkProtocol.V3b\x06proto3')

_POOLTYPE = DESCRIPTOR.enum_types_by_name['PoolType']
PoolType = enum_type_wrapper.EnumTypeWrapper(_POOLTYPE)
_GCPHASE = DESCRIPTOR.enum_types_by_name['GCPhase']
GCPhase = enum_type_wrapper.EnumTypeWrapper(_GCPHASE)
CODE_CACHE_USAGE = 0
NEWGEN_USAGE = 1
OLDGEN_USAGE = 2
SURVIVOR_USAGE = 3
PERMGEN_USAGE = 4
METASPACE_USAGE = 5
NEW = 0
OLD = 1
NORMAL = 2


_JVMMETRICCOLLECTION = DESCRIPTOR.message_types_by_name['JVMMetricCollection']
_JVMMETRIC = DESCRIPTOR.message_types_by_name['JVMMetric']
_MEMORY = DESCRIPTOR.message_types_by_name['Memory']
_MEMORYPOOL = DESCRIPTOR.message_types_by_name['MemoryPool']
_GC = DESCRIPTOR.message_types_by_name['GC']
_THREAD = DESCRIPTOR.message_types_by_name['Thread']
_CLASS = DESCRIPTOR.message_types_by_name['Class']
JVMMetricCollection = _reflection.GeneratedProtocolMessageType('JVMMetricCollection', (_message.Message,), {
  'DESCRIPTOR' : _JVMMETRICCOLLECTION,
  '__module__' : 'language_agent.JVMMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.JVMMetricCollection)
  })
_sym_db.RegisterMessage(JVMMetricCollection)

JVMMetric = _reflection.GeneratedProtocolMessageType('JVMMetric', (_message.Message,), {
  'DESCRIPTOR' : _JVMMETRIC,
  '__module__' : 'language_agent.JVMMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.JVMMetric)
  })
_sym_db.RegisterMessage(JVMMetric)

Memory = _reflection.GeneratedProtocolMessageType('Memory', (_message.Message,), {
  'DESCRIPTOR' : _MEMORY,
  '__module__' : 'language_agent.JVMMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.Memory)
  })
_sym_db.RegisterMessage(Memory)

MemoryPool = _reflection.GeneratedProtocolMessageType('MemoryPool', (_message.Message,), {
  'DESCRIPTOR' : _MEMORYPOOL,
  '__module__' : 'language_agent.JVMMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.MemoryPool)
  })
_sym_db.RegisterMessage(MemoryPool)

GC = _reflection.GeneratedProtocolMessageType('GC', (_message.Message,), {
  'DESCRIPTOR' : _GC,
  '__module__' : 'language_agent.JVMMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.GC)
  })
_sym_db.RegisterMessage(GC)

Thread = _reflection.GeneratedProtocolMessageType('Thread', (_message.Message,), {
  'DESCRIPTOR' : _THREAD,
  '__module__' : 'language_agent.JVMMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.Thread)
  })
_sym_db.RegisterMessage(Thread)

Class = _reflection.GeneratedProtocolMessageType('Class', (_message.Message,), {
  'DESCRIPTOR' : _CLASS,
  '__module__' : 'language_agent.JVMMetric_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.Class)
  })
_sym_db.RegisterMessage(Class)

_JVMMETRICREPORTSERVICE = DESCRIPTOR.services_by_name['JVMMetricReportService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n3org.apache.skywalking.apm.network.language.agent.v3P\001Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\252\002\035SkyWalking.NetworkProtocol.V3'
  _POOLTYPE._serialized_start=1013
  _POOLTYPE._serialized_end=1141
  _GCPHASE._serialized_start=1143
  _GCPHASE._serialized_end=1182
  _JVMMETRICCOLLECTION._serialized_start=70
  _JVMMETRICCOLLECTION._serialized_end=176
  _JVMMETRIC._serialized_start=179
  _JVMMETRIC._serialized_end=430
  _MEMORY._serialized_start=432
  _MEMORY._serialized_end=516
  _MEMORYPOOL._serialized_start=518
  _MEMORYPOOL._serialized_end=629
  _GC._serialized_start=631
  _GC._serialized_end=703
  _THREAD._serialized_start=706
  _THREAD._serialized_end=911
  _CLASS._serialized_start=913
  _CLASS._serialized_end=1010
  _JVMMETRICREPORTSERVICE._serialized_start=1184
  _JVMMETRICREPORTSERVICE._serialized_end=1282
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking.utils.lang import tostring


@tostring
class ProfileTask:

    def __init__(self,
                 task_id: str = '',
                 first_span_op_name: str = '',
                 duration: int = -1,
                 min_duration_threshold: int = -1,
                 thread_dump_period: int = -1,
                 max_sampling_count: int = -1,
                 start_time: int = -1,
                 create_time: int = -1):
        self.task_id = str(task_id)  # type: str
        self.first_span_op_name = str(first_span_op_name)  # type: str
        self.duration = int(duration)  # type: int
        # when can start profile after span context created
        self.min_duration_threshold = int(min_duration_threshold)  # type: int
        # profile interval
        self.thread_dump_period = int(thread_dump_period)  # type: int
        self.max_sampling_count = int(max_sampling_count)  # type: int
        self.start_time = int(start_time)  # type: int
        self.create_time = int(create_time)  # type: int



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from inspect import iscoroutinefunction, isawaitable

from skywalking import Layer, Component, config
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context, NoopContext
from skywalking.trace.span import NoopSpan
from skywalking.trace.tags import TagHttpMethod, TagHttpURL, TagHttpStatusCode

# version_rule = {
#     "name": "tornado",
#     "rules": [">=5.0"]
# }
link_vector = ['https://www.tornadoweb.org']
support_matrix = {
    'tornado': {
        '>=3.6': ['6.0', '6.1']
    }
}
note = """"""


def install():
    from tornado.web import RequestHandler
    old_execute = RequestHandler._execute
    old_log_exception = RequestHandler.log_exception
    RequestHandler._execute = _gen_sw_get_response_func(old_execute)

    def _sw_handler_uncaught_exception(self: RequestHandler, ty, value, tb, *args, **kwargs):
        if value is not None:
            entry_span = get_context().active_span()
            if entry_span is not None:
                entry_span.raised()

        return old_log_exception(self, ty, value, tb, *args, **kwargs)

    RequestHandler.log_exception = _sw_handler_uncaught_exception


def _gen_sw_get_response_func(old_execute):
    from tornado.gen import coroutine

    awaitable = iscoroutinefunction(old_execute)
    if awaitable:
        # Starting Tornado 6 RequestHandler._execute method is a standard Python coroutine (async/await)
        # In that case our method should be a coroutine function too
        async def _sw_get_response(self, *args, **kwargs):
            request = self.request
            carrier = Carrier()
            method = request.method

            for item in carrier:
                if item.key.capitalize() in request.headers:
                    item.val = request.headers[item.key.capitalize()]

            span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
                else get_context().new_entry_span(op=request.path, carrier=carrier)

            with span:
                span.layer = Layer.Http
                span.component = Component.Tornado
                peer = request.connection.stream.socket.getpeername()
                span.peer = f'{peer[0]}:{peer[1]}'
                span.tag(TagHttpMethod(method))
                span.tag(TagHttpURL(f'{request.protocol}://{request.host}{request.path}'))
                result = old_execute(self, *args, **kwargs)
                if isawaitable(result):
                    result = await result
                span.tag(TagHttpStatusCode(self._status_code))
                if self._status_code >= 400:
                    span.error_occurred = True
            return result
    else:
        @coroutine
        def _sw_get_response(self, *args, **kwargs):
            request = self.request
            carrier = Carrier()
            method = request.method

            for item in carrier:
                if item.key.capitalize() in request.headers:
                    item.val = request.headers[item.key.capitalize()]

            span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
                else get_context().new_entry_span(op=request.path, carrier=carrier)

            with span:
                span.layer = Layer.Http
                span.component = Component.Tornado
                peer = request.connection.stream.socket.getpeername()
                span.peer = f'{peer[0]}:{peer[1]}'
                span.tag(TagHttpMethod(method))
                span.tag(TagHttpURL(f'{request.protocol}://{request.host}{request.path}'))
                result = yield from old_execute(self, *args, **kwargs)
                span.tag(TagHttpStatusCode(self._status_code))
                if self._status_code >= 400:
                    span.error_occurred = True
            return result
    return _sw_get_response
